from __future__ import annotations

import collections
import os
import pathlib

import procrunner
import pytest

from cctbx import uctbx
from dxtbx.model import ExperimentList
from dxtbx.serialize import load

from dials.array_family import flex


def unit_cells_are_similar(
    uc1, uc2, relative_length_tolerance=0.01, absolute_angle_tolerance=1
):
    # see also uctbx.cpp unit_cell::is_similar_to()
    l1 = uc1.parameters()
    l2 = uc2.parameters()
    for i in range(3):
        if abs(min(l1[i], l2[i]) / max(l1[i], l2[i]) - 1) > relative_length_tolerance:
            return False
    for i in range(3, 6):
        if abs(l1[i] - l2[i]) > absolute_angle_tolerance:
            if abs(l1[i] - (180 - l2[i])) > absolute_angle_tolerance:
                return False
    return True


_indexing_result = collections.namedtuple(
    "indexing", ["indexed_reflections", "experiments", "rmsds"]
)


def run_indexing(
    reflections,
    experiment,
    working_directory,
    extra_args,
    expected_unit_cell,
    expected_rmsds,
    expected_hall_symbol,
    n_expected_lattices=1,
    relative_length_tolerance=0.005,
    absolute_angle_tolerance=0.5,
):
    commands = ["dials.index"]
    if isinstance(reflections, list):
        commands.extend(reflections)
    else:
        commands.append(reflections)
    if isinstance(experiment, list):
        commands.extend(experiment)
    else:
        commands.append(experiment)
    commands.extend(extra_args)

    result = procrunner.run(commands, working_directory=working_directory)
    assert not result.returncode and not result.stderr

    out_expts = working_directory / "indexed.expt"
    out_refls = working_directory / "indexed.refl"
    assert out_expts.is_file()
    assert out_refls.is_file()

    experiments_list = load.experiment_list(out_expts, check_format=False)
    assert len(experiments_list.crystals()) == n_expected_lattices
    indexed_reflections = flex.reflection_table.from_file(out_refls)
    indexed_reflections.assert_experiment_identifiers_are_consistent(experiments_list)
    rmsds = None

    for i, experiment in enumerate(experiments_list):
        assert unit_cells_are_similar(
            experiment.crystal.get_unit_cell(),
            expected_unit_cell,
            relative_length_tolerance=relative_length_tolerance,
            absolute_angle_tolerance=absolute_angle_tolerance,
        ), (
            experiment.crystal.get_unit_cell().parameters(),
            expected_unit_cell.parameters(),
        )
        sg = experiment.crystal.get_space_group()
        assert sg.type().hall_symbol() == expected_hall_symbol, (
            sg.type().hall_symbol(),
            expected_hall_symbol,
        )
        reflections = indexed_reflections.select(indexed_reflections["id"] == i)
        mi = reflections["miller_index"]
        assert (mi != (0, 0, 0)).count(False) == 0
        reflections = reflections.select(mi != (0, 0, 0))
        reflections = reflections.select(
            reflections.get_flags(reflections.flags.used_in_refinement)
        )
        assert len(reflections) > 0
        obs_x, obs_y, obs_z = reflections["xyzobs.mm.value"].parts()
        calc_x, calc_y, calc_z = reflections["xyzcal.mm"].parts()
        rmsd_x = flex.mean(flex.pow2(obs_x - calc_x)) ** 0.5
        rmsd_y = flex.mean(flex.pow2(obs_y - calc_y)) ** 0.5
        rmsd_z = flex.mean(flex.pow2(obs_z - calc_z)) ** 0.5
        rmsds = (rmsd_x, rmsd_y, rmsd_z)
        for actual, expected in zip(rmsds, expected_rmsds):
            assert actual <= expected, f"{rmsds} {expected_rmsds}"
        assert experiment.identifier != ""
        expt = ExperimentList()
        expt.append(experiment)
        reflections.assert_experiment_identifiers_are_consistent(expt)

    return _indexing_result(indexed_reflections, experiments_list, rmsds)


def test_index_i04_weak_data_fft3d(dials_regression, tmp_path):
    # thaumatin
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "full.pickle")
    sequence_path = os.path.join(data_dir, "experiments_import.json")
    extra_args = [
        "bin_size_fraction=0.25",
        "image_range=1,20",
        "image_range=250,270",
        "image_range=520,540",
    ]
    expected_unit_cell = uctbx.unit_cell((57.7, 57.7, 149.8, 90, 90, 90))
    expected_rmsds = (0.05, 0.04, 0.0005)
    expected_hall_symbol = " P 1"

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_index_trypsin_four_lattice_P212121(dials_regression, tmp_path):
    # synthetic trypsin multi-lattice dataset (4 lattices)
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "trypsin"
    pickle_path = data_dir / "P1_X6_1_2_3_4.pickle"
    sequence_path = data_dir / "experiments_P1_X6_1_2_3_4.json"
    extra_args = [
        "indexing.method=real_space_grid_search",
        "reflections_per_degree=10",
        "n_macro_cycles=5",
        "known_symmetry.unit_cell=54.3,58.3,66.5,90,90,90",
        "known_symmetry.space_group=P212121",
        "image_range=0,10",
        "beam.fix=all",
        "detector.fix=all",
        "max_cell=70",
    ]
    expected_unit_cell = uctbx.unit_cell((54.3, 58.3, 66.5, 90, 90, 90))
    expected_rmsds = (0.28, 0.30, 0.006)
    expected_hall_symbol = " P 2ac 2ab"
    n_expected_lattices = 1

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
        n_expected_lattices=n_expected_lattices,
        relative_length_tolerance=0.02,
        absolute_angle_tolerance=1,
    )


def test_index_i04_weak_data_fft1d(dials_regression, tmp_path):
    # thaumatin
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "i04_weak_data"
    pickle_path = data_dir / "full.pickle"
    sequence_path = data_dir / "experiments_import.json"
    extra_args = [
        "n_macro_cycles=2",
        "indexing.method=fft1d",
        "bin_size_fraction=0.25",
        "image_range=1,20",
        "image_range=250,270",
        "image_range=520,540",
    ]
    expected_unit_cell = uctbx.unit_cell((57.7, 57.7, 149.9, 90, 90, 90))
    expected_rmsds = (0.06, 0.05, 0.0005)
    expected_hall_symbol = " P 1"

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_index_trypsin_index_assignment_local(dials_regression, tmp_path):
    # synthetic trypsin multi-lattice dataset (3 lattices)
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "trypsin"
    pickle_path = data_dir / "P1_X6_1_2_3.pickle"
    sequence_path = data_dir / "experiments_P1_X6_1_2_3.json"
    extra_args = [
        "indexing.method=real_space_grid_search",
        "d_min_start=3",
        "n_macro_cycles=3",
        "known_symmetry.unit_cell=54.3,58.3,66.5,90,90,90",
        "known_symmetry.space_group=P212121",
        "image_range=0,10",
        "beam.fix=all",
        "detector.fix=all",
        "max_lattices=3",
        "index_assignment.method=local",
        "nearest_neighbours=50",
    ]

    expected_unit_cell = uctbx.unit_cell((54.3, 58.3, 66.5, 90, 90, 90))
    expected_rmsds = (0.33, 0.40, 0.0024)
    expected_hall_symbol = " P 2ac 2ab"
    n_expected_lattices = 3

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
        n_expected_lattices=n_expected_lattices,
        relative_length_tolerance=0.02,
        absolute_angle_tolerance=1,
    )


def test_index_peak_search_clean(dials_regression, tmp_path):
    # test indexing from single image of i04_weak_data
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "i04_weak_data"
    pickle_path = data_dir / "first_image.pickle"
    sequence_path = data_dir / "experiments_import.json"
    extra_args = [
        "indexing.method=fft3d",
        "known_symmetry.space_group=P4",
        "known_symmetry.unit_cell=57.8,57.8,150,90,90,90",
        "peak_search=clean",
        "min_samples=15",
        "n_macro_cycles=4",
        "reciprocal_space_grid.d_min=4",
    ]

    expected_unit_cell = uctbx.unit_cell((57.8, 57.8, 150, 90, 90, 90))
    expected_rmsds = (0.06, 0.07, 0.003)
    expected_hall_symbol = " P 4"

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


@pytest.mark.parametrize("specify_unit_cell", [False, True])
def test_index_imosflm_tutorial(dials_regression, tmp_path, specify_unit_cell):
    # test on spots derived from imosflm tutorial data:
    # http://www.ccp4.ac.uk/courses/BCA2005/tutorials/dataproc-tutorial.html
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "imosflm_hg_mar"
    pickle_path = data_dir / "strong.pickle"
    sequence_path = data_dir / "experiments.json"

    unit_cell = uctbx.unit_cell((58.373, 58.373, 155.939, 90, 90, 120))
    hall_symbol = '-R 3 2"'

    extra_args = [
        "bin_size_fraction=0.25",
        'known_symmetry.space_group="Hall: %s"' % hall_symbol.replace('"', '\\"'),
    ]
    if specify_unit_cell:
        extra_args.append(
            'known_symmetry.unit_cell="%s %s %s %s %s %s"' % unit_cell.parameters()
        )

    expected_unit_cell = unit_cell
    expected_hall_symbol = hall_symbol
    expected_rmsds = (0.08, 0.11, 0.004)

    run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


@pytest.fixture(scope="session")
def insulin_spotfinding(dials_data, tmp_path_factory):
    """Return experiment and reflection files for 2 images of the insulin dataset"""

    data_dir = dials_data("insulin", pathlib=True)
    tmp_path = tmp_path_factory.mktemp("insulin")

    command = ["dials.import"]
    for i, image_path in enumerate(("insulin_1_001.img", "insulin_1_045.img")):
        command.append(data_dir / image_path)

    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    experiment = tmp_path / "imported.expt"
    assert experiment.is_file()

    command = ["dials.find_spots", "nproc=1", experiment]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    reflections = tmp_path / "strong.refl"
    assert reflections.is_file()

    return experiment, reflections


@pytest.mark.parametrize("method", ["fft3d", "fft1d", "real_space_grid_search"])
def test_index_insulin_multi_sequence(insulin_spotfinding, tmp_path, method):
    experiment, reflections = insulin_spotfinding
    expected_unit_cell = uctbx.unit_cell(
        (78.163, 78.163, 78.163, 90.000, 90.000, 90.000)
    )
    expected_hall_symbol = " I 2 2 3"
    expected_rmsds = (0.05, 0.06, 0.01)
    extra_args = [
        'known_symmetry.unit_cell="%s %s %s %s %s %s"'
        % expected_unit_cell.parameters(),
        f'known_symmetry.space_group="Hall: {expected_hall_symbol}"',
        f"indexing.method={method}",
        "treat_single_image_as_still=False",
    ]
    run_indexing(
        reflections,
        experiment,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


@pytest.fixture(scope="session")
def insulin_spotfinding_stills(dials_data, tmp_path_factory):
    """Return experiment and reflection files for 1 image of the insulin
    dataset treated as still image"""

    data_dir = dials_data("insulin", pathlib=True)
    tmp_path = tmp_path_factory.mktemp("insulin")

    command = [
        "dials.import",
        "convert_sequences_to_stills=True",
        data_dir / "insulin_1_001.img",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    experiment = tmp_path / "imported.expt"
    assert experiment.is_file()

    command = ["dials.find_spots", "nproc=1", experiment]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    reflections = tmp_path / "strong.refl"
    assert reflections.is_file()

    return experiment, reflections


@pytest.mark.xfel
@pytest.mark.parametrize("method", ["fft3d", "fft1d", "real_space_grid_search"])
def test_index_insulin_force_stills(insulin_spotfinding_stills, tmp_path, method):
    experiment, reflections = insulin_spotfinding_stills
    expected_unit_cell = uctbx.unit_cell(
        (78.092, 78.092, 78.092, 90.000, 90.000, 90.000)
    )
    expected_hall_symbol = " I 2 2 3"
    expected_rmsds = (0.05, 0.06, 0.01)

    extra_args = [
        "stills.indexer=stills",
        'known_symmetry.unit_cell="%s %s %s %s %s %s"'
        % expected_unit_cell.parameters(),
        f'known_symmetry.space_group="Hall: {expected_hall_symbol}"',
        f"indexing.method={method}",
    ]

    run_indexing(
        reflections,
        experiment,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_multiple_experiments(dials_regression, tmp_path):
    # Test indexing 4 lysozyme still shots in a single dials.index job
    #   - the first image doesn't index
    #   - the last three images do index
    data_dir = (
        pathlib.Path(dials_regression) / "indexing_test_data" / "i24_lysozyme_stills"
    )
    pickle_path = data_dir / "strong.pickle"
    experiments_json = data_dir / "imported_experiments.json"

    expected_unit_cell = uctbx.unit_cell((38.06, 78.78, 78.91, 90, 90, 90))
    expected_hall_symbol = " P 1"
    expected_rmsds = (0.1, 0.07, 0.0)

    extra_args = [
        "stills.indexer=sequences",
        "joint_indexing=False",
        "outlier.algorithm=sauter_poon",
    ]

    run_indexing(
        pickle_path,
        experiments_json,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
        n_expected_lattices=3,
        relative_length_tolerance=0.01,
    )


def test_index_4rotation(dials_regression, tmp_path):
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "4rotation"
    pickle_path = data_dir / "strong.pickle"
    sequence_path = data_dir / "experiments.json"
    extra_args = [
        "max_refine=10",
        "reflections_per_degree=50",
        "known_symmetry.space_group=R3",
        "n_macro_cycles=3",
    ]
    expected_unit_cell = uctbx.unit_cell((48.397, 48.397, 284.767, 90, 90, 120))
    expected_rmsds = (0.06, 0.08, 0.22)
    expected_hall_symbol = " R 3"

    result = run_indexing(
        pickle_path,
        sequence_path,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )
    assert len(result.indexed_reflections) > 276800, len(result.indexed_reflections)


def test_index_small_molecule_multi_sequence_4(dials_regression, tmp_path):
    # test for small molecule multi-sequence indexing, 4 sequences with different values
    # of goniometer.fixed_rotation()
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "multi_sweep"
    pickle_paths = [
        sorted((data_dir / f"SWEEP{i + 1}" / "index").glob("*_strong.pickle"))[0]
        for i in range(4)
    ]
    sequence_paths = [
        data_dir / f"SWEEP{i + 1}" / "index" / "experiments.json" for i in range(4)
    ]
    extra_args = ["known_symmetry.space_group=I4", "filter_ice=False"]
    expected_unit_cell = uctbx.unit_cell((7.310, 7.310, 6.820, 90.000, 90.000, 90.000))
    expected_rmsds = (0.10, 0.7, 0.5)
    expected_hall_symbol = " I 4"

    result = run_indexing(
        pickle_paths,
        sequence_paths,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )
    assert len(result.indexed_reflections) > 1250, len(result.indexed_reflections)


def test_index_small_molecule_multi_sequence_3(dials_regression, tmp_path):
    # test for small molecule multi-sequence indexing, 3 sequences with different values
    # of goniometer setting rotation (i.e. phi scans)
    data_dir = pathlib.Path(dials_regression) / "dials-191"
    print(data_dir)
    pickle_paths = [
        sorted(data_dir.glob(f"*_SWEEP{i + 1}_strong.pickle"))[0] for i in range(3)
    ]
    sequence_paths = [
        sorted(data_dir.glob(f"*_SWEEP{i + 1}_experiments.json"))[0] for i in range(3)
    ]
    extra_args = ["filter_ice=False"]
    expected_unit_cell = uctbx.unit_cell(
        (9.440, 15.313, 17.126, 90.073, 90.106, 79.248)
    )
    expected_rmsds = (0.32, 0.34, 0.005)
    expected_hall_symbol = " P 1"

    result = run_indexing(
        pickle_paths,
        sequence_paths,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )
    assert len(result.indexed_reflections) > 12000, len(result.indexed_reflections)
    # expect at least indexed 2000 reflections per experiment
    for i in range(3):
        assert (result.indexed_reflections["id"] == i).count(True) > 2000


def test_index_small_molecule_ice_max_cell(dials_regression, tmp_path):
    # test for small molecule indexing: presence of ice rings makes max-cell
    # estimation tricky
    data_dir = os.path.join(dials_regression, "indexing_test_data", "MXSW-904")
    pickle_path = os.path.join(data_dir, "1_SWEEP1_strong.pickle")
    experiments = os.path.join(data_dir, "1_SWEEP1_experiments.json")
    extra_args = ["filter_ice=False"]
    expected_unit_cell = uctbx.unit_cell((11.72, 11.72, 11.74, 109.08, 109.24, 108.99))
    expected_rmsds = (0.06, 0.05, 0.04)
    expected_hall_symbol = " P 1"

    result = run_indexing(
        pickle_path,
        experiments,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )
    assert len(result.indexed_reflections) > 1300, len(result.indexed_reflections)


@pytest.mark.xfail
def test_refinement_failure_on_max_lattices_a15(dials_regression, tmp_path):
    """Problem: Sometimes there is enough data to index, but not enough to
    refine. If this happens in the (N>1)th crystal of max_lattices, then
    all existing solutions are also dropped."""
    data_dir = os.path.join(dials_regression, "indexing_test_data", "lattice_failures")

    result = procrunner.run(
        [
            "dials.index",
            os.path.join(data_dir, "lpe4-2-a15_strong.pickle"),
            os.path.join(data_dir, "lpe4-2-a15_datablock.json"),
            "max_lattices=3",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.refl").if_file()
    assert (tmp_path / "indexed.expt").is_file()
    experiments_list = load.experiment_list(
        tmp_path / "indexed.expt", check_format=False
    )
    assert len(experiments_list) == 2

    # now try to reindex with existing model
    result = procrunner.run(
        [
            "dials.index",
            tmp_path / "indexed.expt",
            os.path.join(data_dir, "lpe4-2-a15_strong.pickle"),
            "max_lattices=2",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.refl").is_file()
    assert (tmp_path / "indexed.expt").is_file()
    experiments_list = load.experiment_list(
        tmp_path / "indexed.expt", check_format=False
    )
    assert len(experiments_list) == 2


@pytest.mark.xfel
def test_stills_indexer_multi_lattice_bug_MosaicSauter2014(dials_regression, tmp_path):
    """Problem: In stills_indexer, before calling the refine function, the
    experiment list contains a list of dxtbx crystal models (that are not
    MosaicSauter2014 models). The conversion to MosaicSauter2014 is made
    during the refine step when functions from nave_parameters is called.
    If the experiment list contains more than 1 experiment, for eg.
    multiple lattices, only the first crystal gets assigned mosaicity. In
    actuality, all crystal models should be assigned mosaicity. This test
    only compares whether or not all crystal models have been assigned a
    MosaicSauter2014 model."""

    import dxtbx.model
    from dxtbx.model import Crystal
    from dxtbx.model.experiment_list import (
        Experiment,
        ExperimentList,
        ExperimentListFactory,
    )

    from dials.algorithms.indexing.stills_indexer import StillsIndexer
    from dials.array_family import flex
    from dials.command_line.stills_process import (
        phil_scope as stills_process_phil_scope,
    )

    experiment_data = os.path.join(
        dials_regression,
        "refinement_test_data",
        "cspad_refinement",
        "cspad_refined_experiments_step6_level2_300.json",
    )
    reflection_data = os.path.join(
        dials_regression,
        "refinement_test_data",
        "cspad_refinement",
        "cspad_reflections_step7_300.pickle",
    )

    refl = flex.reflection_table.from_file(reflection_data)
    explist = ExperimentListFactory.from_json_file(experiment_data, check_format=False)[
        0:2
    ]
    reflist = refl.select(refl["id"] < 2)  # Only use the first 2 for convenience
    # Construct crystal models that don't have mosaicity. These A,B,C values are the same
    # as read in from the dials_regression folder
    # Crystal-0
    cs0 = Crystal(explist[0].crystal)
    exp0 = Experiment(
        imageset=explist[0].imageset,
        beam=explist[0].beam,
        detector=explist[0].detector,
        goniometer=None,
        scan=None,
        crystal=cs0,
    )

    # Crystal-1
    cs1 = Crystal(explist[1].crystal)
    exp1 = Experiment(
        imageset=explist[1].imageset,
        beam=explist[1].beam,
        detector=explist[1].detector,
        goniometer=None,
        scan=None,
        crystal=cs1,
    )
    # Construct a new experiment_list that will be passed on for refinement
    unrefined_explist = ExperimentList([exp0, exp1])
    # Get default params from stills_process and construct StillsIndexer, then run refinement
    params = stills_process_phil_scope.extract()
    SI = StillsIndexer(reflist, unrefined_explist, params=params)
    refined_explist, new_reflist = SI.refine(unrefined_explist, reflist)
    # Now check whether the models have mosaicity after stills_indexer refinement
    # Also check that mosaicity values are within expected limits
    for ii, crys in enumerate(refined_explist.crystals()):
        assert isinstance(crys, dxtbx.model.MosaicCrystalSauter2014)
        if ii == 0:
            assert crys.get_domain_size_ang() == pytest.approx(2242.0, rel=0.1)
        if ii == 1:
            assert crys.get_domain_size_ang() == pytest.approx(2689.0, rel=0.1)


@pytest.mark.parametrize(
    "indexer_type,fix_cell",
    (("sequences", False), pytest.param("stills", True, marks=pytest.mark.xfel)),
)
def test_index_ED_still_low_res_spot_match(
    dials_data, tmp_path, indexer_type, fix_cell
):
    # test indexing from a single simulated lysozyme ED still

    image_path = (
        dials_data("image_examples", pathlib=True) / "simtbx_FormatSMVJHSim_001.img"
    )

    command = ["dials.import", image_path]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    experiment = tmp_path / "imported.expt"
    assert experiment.is_file()

    command = ["dials.find_spots", "nproc=1", experiment]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr

    reflections = tmp_path / "strong.refl"

    extra_args = [
        "indexing.method=low_res_spot_match",
        "known_symmetry.space_group=P43212",
        "known_symmetry.unit_cell=78.84,78.84,38.29,90,90,90",
        "stills.indexer=" + indexer_type,
        "n_macro_cycles=2",
        "detector.fix_list=Dist",
    ]
    if fix_cell:
        extra_args += ["crystal.fix=cell"]

    expected_unit_cell = uctbx.unit_cell((78.84, 78.84, 38.29, 90, 90, 90))
    expected_rmsds = (0.0065, 0.0065, 0.000)
    expected_hall_symbol = " P 4nw 2abw"

    run_indexing(
        reflections,
        experiment,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


@pytest.mark.parametrize(
    "cell_params",
    [
        (44.47, 52.85, 62.23, 115.14, 101.72, 90.01),
        (52.85, 62.23, 44.47, 101.72, 90.01, 115.14),
    ],
)
def test_unconventional_P1_cell(dials_data, tmp_path, cell_params):
    """
    Indexing in P1 should succeed even if the cell parameters are provided in
    a non-conventional setting
    """
    data_dir = dials_data("mpro_x0305_processed", pathlib=True)
    experiment = data_dir / "imported.expt"
    reflections = data_dir / "strong.refl"

    cell_params_str = ",".join([str(x) for x in cell_params])
    extra_args = [
        "indexing.method=fft3d",
        "known_symmetry.space_group=P1",
        "known_symmetry.unit_cell=" + cell_params_str,
    ]
    expected_unit_cell = uctbx.unit_cell(cell_params)
    expected_rmsds = (1, 1, 1)
    expected_hall_symbol = " P 1"

    run_indexing(
        reflections,
        experiment,
        tmp_path,
        extra_args,
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_real_space_grid_search_no_unit_cell(dials_regression, tmp_path):
    data_dir = pathlib.Path(dials_regression) / "indexing_test_data" / "i04_weak_data"
    experiments_json = data_dir / "experiments_import.json"
    pickle_path = data_dir / "full.pickle"
    commands = [
        "dials.index",
        experiments_json,
        pickle_path,
        "indexing.method=real_space_grid_search",
    ]
    result = procrunner.run(commands, working_directory=tmp_path)
    assert result.stderr
    assert (
        result.stderr.strip()
        == b"Target unit cell must be provided for real_space_grid_search"
    )


def test_index_known_orientation(dials_data, tmp_path):
    data_dir = dials_data("vmxi_proteinase_k_sweeps", pathlib=True)
    experiments_json = data_dir / "experiments_0.json"
    reflections = data_dir / "reflections_0.pickle"

    expected_unit_cell = uctbx.unit_cell((68.395, 68.395, 104, 90, 90, 90))
    expected_rmsds = (0.013, 0.012, 0.008)
    expected_hall_symbol = " P 4"

    run_indexing(
        reflections,
        experiments_json,
        tmp_path,
        [],
        expected_unit_cell,
        expected_rmsds,
        expected_hall_symbol,
    )


def test_all_expt_ids_have_expts(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.index",
            dials_data("vmxi_thaumatin_grid_index", pathlib=True) / "split_07602.expt",
            dials_data("vmxi_thaumatin_grid_index", pathlib=True) / "split_07602.refl",
            "stills.indexer=sequences",
            "indexing.method=real_space_grid_search",
            "space_group=P4",
            "unit_cell=58,58,150,90,90,90",
            "max_lattices=8",
            "beam.fix=all",
            "detector.fix=all",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert (tmp_path / "indexed.expt").is_file()
    assert (tmp_path / "indexed.refl").is_file()

    refl = flex.reflection_table.from_file(tmp_path / "indexed.refl")
    expt = ExperimentList.from_file(tmp_path / "indexed.expt", check_format=False)

    assert flex.max(refl["id"]) + 1 == len(expt)



from __future__ import annotations

import pickle

import procrunner


def return_locations():
    locations = {
        (1042, 980),
        (1042, 988),
        (1042, 1474),
        (1060, 980),
        (1060, 988),
        (1060, 1474),
        (1060, 1482),
        (1253, 980),
        (1254, 980),
        (1254, 988),
        (1254, 1474),
        (1254, 1482),
        (1272, 1474),
        (1272, 1482),
        (1285, 1235),
        (1285, 1237),
        (1465, 980),
        (1466, 980),
        (1466, 988),
        (1466, 1474),
        (1484, 988),
        (1484, 1482),
        (1696, 1482),
    }
    return locations


def test_find_bad_pixels(dials_data, tmp_path):

    image_files = sorted(dials_data("x4wide", pathlib=True).glob("*.cbf"))
    image_files = image_files[:10] + image_files[-10:]
    result = procrunner.run(
        [
            "dials.find_bad_pixels",
            "mask=pixels.mask",
        ]
        + image_files,
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    locations = return_locations()
    # verify bad pixels
    for record in result.stdout.decode().split("\n"):
        if "mask" in record and "Wrote" not in record:
            tokens = tuple(map(int, record.split()[1:]))
            assert tokens[-1] == 16
            position = tokens[:2]
            locations.remove(position)

    assert len(locations) == 0

    assert (tmp_path / "pixels.mask").is_file()
    with open(tmp_path / "pixels.mask", "rb") as fh:
        mask = pickle.load(fh, encoding="bytes")

    # X4-wide image size = 2463, 2527
    n_fast = 2463
    locations = return_locations()
    # locations are reported slow fast

    assert mask.count(False) == len(locations)
    for loc in locations:
        idx = (loc[0] * n_fast) + loc[1]
        assert mask[idx] is False

    # check that this mask can be used in import
    image_files = sorted(dials_data("x4wide", pathlib=True).glob("*.cbf"))[:3]
    result = procrunner.run(
        [
            "dials.import",
            "mask=pixels.mask",
        ]
        + image_files,
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr



from __future__ import annotations

from dials.algorithms.refinement.outlier_detection import CentroidOutlier


class SauterPoon(CentroidOutlier):
    """Implementation of the CentroidOutlier class using the algorithm of
    Sauter & Poon (2010) (https://doi.org/10.1107/S0021889810010782)."""

    def __init__(
        self,
        cols=None,
        min_num_obs=20,
        separate_experiments=True,
        separate_panels=True,
        separate_images=False,
        block_width=None,
        nproc=1,
        px_sz=(1, 1),
        verbose=False,
        pdf=None,
    ):

        # here the column names are fixed by the algorithm, so what's passed in is
        # ignored.
        CentroidOutlier.__init__(
            self,
            cols=["miller_index", "xyzobs.px.value", "xyzcal.px"],
            min_num_obs=min_num_obs,
            separate_experiments=separate_experiments,
            separate_panels=separate_panels,
            separate_images=separate_images,
            block_width=block_width,
            nproc=nproc,
        )

        self._px_sz = px_sz
        self._verbose = verbose
        self._pdf = pdf

        return

    def _detect_outliers(self, cols):

        # cols is guaranteed to be a list of three flex arrays, containing miller
        # indices, observed pixel coordinates and calculated pixel coordinates.
        # Copy the data into matches
        class match:
            pass

        matches = []
        for hkl in cols[0]:
            m = match()
            m.miller_index = hkl
            matches.append(m)

        for obs, m in zip(cols[1], matches):
            m.x_obs = obs[0] * self._px_sz[0]
            m.y_obs = obs[1] * self._px_sz[1]

        for calc, m in zip(cols[2], matches):
            m.x_calc = calc[0] * self._px_sz[0]
            m.y_calc = calc[1] * self._px_sz[1]

        import iotbx.phil
        from rstbx.phil.phil_preferences import indexing_api_defs

        hardcoded_phil = iotbx.phil.parse(input_string=indexing_api_defs).extract()

        # set params into the hardcoded_phil
        hardcoded_phil.indexing.outlier_detection.verbose = self._verbose
        hardcoded_phil.indexing.outlier_detection.pdf = self._pdf

        from rstbx.indexing_api.outlier_procedure import OutlierPlotPDF

        if self._pdf is not None:
            ## new code for outlier rejection inline here
            hardcoded_phil.__inject__(
                "writer", OutlierPlotPDF(hardcoded_phil.indexing.outlier_detection.pdf)
            )

        # execute Sauter and Poon (2010) algorithm
        from rstbx.indexing_api import outlier_detection

        od = outlier_detection.find_outliers_from_matches(
            matches, verbose=self._verbose, horizon_phil=hardcoded_phil
        )

        # flex.bool of the inliers
        outliers = ~od.get_cache_status()

        return outliers



from __future__ import annotations

import collections
import os
import re

import libtbx.phil
from dxtbx.model.experiment_list import ExperimentListFactory
from libtbx.utils import Sorry

from dials.array_family import flex

FilenameDataWrapper = collections.namedtuple("FilenameDataWrapper", "filename, data")


class ExperimentListConverters:
    """A phil converter for the experiment list class."""

    phil_type = "experiment_list"

    def __init__(self, check_format=True):
        self._check_format = check_format

    def __str__(self):
        return self.phil_type

    def from_words(self, words, master):
        s = libtbx.phil.str_from_words(words=words)
        if s is None:
            return None
        if s == "<image files>":
            return FilenameDataWrapper(filename=s, data=None)
        if not os.path.exists(s):
            raise Sorry(f"File {s} does not exist")
        return FilenameDataWrapper(
            filename=s,
            data=ExperimentListFactory.from_json_file(
                s, check_format=self._check_format
            ),
        )

    def as_words(self, python_object, master):
        if python_object is None:
            value = "None"
        else:
            value = python_object.filename
        return [libtbx.phil.tokenizer.word(value=value)]


class ReflectionTableConverters:
    """A phil converter for the reflection table class."""

    phil_type = "reflection_table"

    def __str__(self):
        return self.phil_type

    def from_words(self, words, master):
        s = libtbx.phil.str_from_words(words=words)
        if s is None:
            return None
        if not os.path.exists(s):
            raise Sorry(f"File {s} does not exist")
        return FilenameDataWrapper(filename=s, data=flex.reflection_table.from_file(s))

    def as_words(self, python_object, master):
        if python_object is None:
            value = "None"
        else:
            value = python_object.filename
        return [libtbx.phil.tokenizer.word(value=value)]


class ReflectionTableSelectorConverters:
    """A phil converter for the reflection table selector class."""

    phil_type = "reflection_table_selector"

    def __str__(self):
        return self.phil_type

    def from_words(self, words, master):
        s = libtbx.phil.str_from_words(words=words)
        if s is None:
            return None
        regex = r"^\s*([\w\.]+)\s*(<=|!=|==|>=|<|>|&)\s*(.+)\s*$"
        matches = re.findall(regex, s)
        if len(matches) == 0:
            raise RuntimeError(f"{s} is not a valid selection")
        elif len(matches) > 1:
            raise RuntimeError(f"{s} is not a single selection")
        else:
            match = matches[0]
        assert len(match) == 3
        col, op, value = match
        return flex.reflection_table_selector(col, op, value)

    def as_words(self, python_object, master):
        if python_object is None:
            value = "None"
        else:
            value = "{}{}{}".format(
                python_object.column,
                python_object.op_string,
                python_object.value,
            )
        return [libtbx.phil.tokenizer.word(value=value)]


# Create the default converter registry with the extract converters
default_converter_registry = libtbx.phil.extended_converter_registry(
    additional_converters=[
        ExperimentListConverters,
        ReflectionTableConverters,
        ReflectionTableSelectorConverters,
    ]
)


def parse(
    input_string=None,
    source_info=None,
    file_name=None,
    converter_registry=None,
    process_includes=False,
):
    """Redefinition of the parse function."""
    if converter_registry is None:
        converter_registry = default_converter_registry
    return libtbx.phil.parse(
        input_string=input_string,
        source_info=source_info,
        file_name=file_name,
        converter_registry=converter_registry,
        process_includes=process_includes,
    )



from __future__ import annotations

import sys

from dxtbx.model.experiment_list import ExperimentList
from scitbx.matrix import col

"""
Example script for rotating a detector in an experiment.json
"""

experiments = ExperimentList.from_file(sys.argv[1])
print("Original detector")
detector = experiments[0].detector
print(detector)

h = detector.hierarchy()

fast = col(h.get_fast_axis())
slow = col(h.get_slow_axis())
origin = col(h.get_origin())

normal = col(h.get_normal())
rotation = normal.axis_and_angle_as_r3_rotation_matrix(90, deg=True)

h.set_frame(rotation * fast, rotation * slow, rotation * origin)

print("Rotated detector")
print(detector)

experiments.as_json("rotated.expt")



from __future__ import annotations

from dials.array_family import flex
from dials.array_family.flex import Binner
from dials.constants import FULL_PARTIALITY
from dials.util.report import Array, Report, Table


def flex_ios(val, var):
    """
    Compute I/sigma or return zero for each element.
    """
    assert len(val) == len(var)
    result = flex.double(len(val), 0)
    indices = flex.size_t(range(len(val))).select(var > 0)
    val = val.select(indices)
    var = var.select(indices)
    assert var.all_gt(0)
    result.set_selected(indices, val / flex.sqrt(var))
    return result


def generate_integration_report(experiment, reflections, n_resolution_bins=20):
    """
    Generate the integration report
    """
    from cctbx import crystal, miller

    from dials.algorithms.statistics import (
        pearson_correlation_coefficient,
        spearman_correlation_coefficient,
    )

    def overall_report(data):
        # Start by adding some overall numbers
        report = {
            "n": len(reflections),
            "n_full": data["full"].count(True),
            "n_partial": data["full"].count(False),
            "n_overload": data["over"].count(True),
            "n_ice": data["ice"].count(True),
            "n_summed": data["sum"].count(True),
            "n_fitted": data["prf"].count(True),
            "n_integated": data["int"].count(True),
            "n_invalid_bg": data["ninvbg"].count(True),
            "n_invalid_fg": data["ninvfg"].count(True),
            "n_failed_background": data["fbgd"].count(True),
            "n_failed_summation": data["fsum"].count(True),
            "n_failed_fitting": data["fprf"].count(True),
        }

        # Compute mean background
        try:
            report["mean_background"] = flex.mean(
                data["background.mean"].select(data["int"])
            )
        except Exception:
            report["mean_background"] = 0.0

        # Compute mean I/Sigma summation
        try:
            report["ios_sum"] = flex.mean(data["intensity.sum.ios"].select(data["sum"]))
        except Exception:
            report["ios_sum"] = 0.0

        # Compute mean I/Sigma profile fitting
        try:
            report["ios_prf"] = flex.mean(data["intensity.prf.ios"].select(data["prf"]))
        except Exception:
            report["ios_prf"] = 0.0

        # Compute the mean profile correlation
        try:
            report["cc_prf"] = flex.mean(
                data["profile.correlation"].select(data["prf"])
            )
        except Exception:
            report["cc_prf"] = 0.0

        # Compute the correlations between summation and profile fitting
        try:
            mask = data["sum"] & data["prf"]
            Isum = data["intensity.sum.value"].select(mask)
            Iprf = data["intensity.prf.value"].select(mask)
            report["cc_pearson_sum_prf"] = pearson_correlation_coefficient(Isum, Iprf)
            report["cc_spearman_sum_prf"] = spearman_correlation_coefficient(Isum, Iprf)
        except Exception:
            report["cc_pearson_sum_prf"] = 0.0
            report["cc_spearman_sum_prf"] = 0.0

        # Return the overall report
        return report

    def binned_report(binner, index, data):

        # Create the indexers
        indexer_all = binner.indexer(index)
        indexer_sum = binner.indexer(index.select(data["sum"]))
        indexer_prf = binner.indexer(index.select(data["prf"]))
        indexer_int = binner.indexer(index.select(data["int"]))

        # Add some stats by resolution
        report = {
            "bins": list(binner.bins()),
            "n_full": list(indexer_all.sum(data["full"])),
            "n_partial": list(indexer_all.sum(~data["full"])),
            "n_overload": list(indexer_all.sum(data["over"])),
            "n_ice": list(indexer_all.sum(data["ice"])),
            "n_summed": list(indexer_all.sum(data["sum"])),
            "n_fitted": list(indexer_all.sum(data["prf"])),
            "n_integrated": list(indexer_all.sum(data["int"])),
            "n_invalid_bg": list(indexer_all.sum(data["ninvbg"])),
            "n_invalid_fg": list(indexer_all.sum(data["ninvfg"])),
            "n_failed_background": list(indexer_all.sum(data["fbgd"])),
            "n_failed_summation": list(indexer_all.sum(data["fsum"])),
            "n_failed_fitting": list(indexer_all.sum(data["fprf"])),
        }

        # Compute mean background
        try:
            report["mean_background"] = list(
                indexer_int.mean(data["background.mean"].select(data["int"]))
            )
        except Exception:
            report["mean_background"] = [0.0] * len(binner)

        # Compute mean I/Sigma summation
        try:
            report["ios_sum"] = list(
                indexer_sum.mean(data["intensity.sum.ios"].select(data["sum"]))
            )
        except Exception:
            report["ios_sum"] = [0.0] * len(binner)

        # Compute mean I/Sigma profile fitting
        try:
            report["ios_prf"] = list(
                indexer_prf.mean(data["intensity.prf.ios"].select(data["prf"]))
            )
        except Exception:
            report["ios_prf"] = [0.0] * len(binner)

        # Compute the mean profile correlation
        try:
            report["cc_prf"] = list(
                indexer_prf.mean(data["profile.correlation"].select(data["prf"]))
            )
        except Exception:
            report["cc_prf"] = [0.0] * len(binner)

        try:
            report["rmsd_xy"] = list(
                indexer_sum.mean(data["xyz.rmsd"].select(data["sum"]))
            )
        except Exception:
            report["rmsd_xy"] = [0.0] * len(binner)

        # Return the binned report
        return report

    def resolution_bins(experiment, hkl, nbins):

        # Create the crystal symmetry object
        cs = crystal.symmetry(
            space_group=experiment.crystal.get_space_group(),
            unit_cell=experiment.crystal.get_unit_cell(),
        )

        # Create the resolution binner object
        ms = miller.set(cs, hkl)
        ms.setup_binner(n_bins=nbins)
        binner = ms.binner()
        brange = list(binner.range_used())
        bins = [binner.bin_d_range(brange[0])[0]]
        for i in brange:
            bins.append(binner.bin_d_range(i)[1])
        return flex.double(reversed(bins))

    def select(data, indices):
        # Select rows from columns
        result = {key: value.select(indices) for key, value in data.items()}
        return result

    # Check the required columns are there
    assert "miller_index" in reflections
    assert "d" in reflections
    assert "flags" in reflections
    assert "bbox" in reflections
    assert "xyzcal.px" in reflections
    assert "partiality" in reflections
    assert "intensity.sum.value" in reflections
    assert "intensity.sum.variance" in reflections

    # Get the flag enumeration
    flags = flex.reflection_table.flags

    # Get some keys from the data
    data = {}
    for key in [
        "miller_index",
        "xyzcal.px",
        "xyzobs.px.value",
        "d",
        "bbox",
        "background.mean",
        "partiality",
        "intensity.sum.value",
        "intensity.sum.variance",
        "intensity.prf.value",
        "intensity.prf.variance",
        "profile.correlation",
    ]:
        if key in reflections:
            data[key] = reflections[key]

    # Compute some flag stuff
    data["full"] = data["partiality"] >= FULL_PARTIALITY
    data["over"] = reflections.get_flags(flags.overloaded)
    data["ice"] = reflections.get_flags(flags.in_powder_ring)
    data["sum"] = reflections.get_flags(flags.integrated_sum)
    data["prf"] = reflections.get_flags(flags.integrated_prf)
    data["int"] = reflections.get_flags(flags.integrated, all=False)
    data["ninvbg"] = reflections.get_flags(flags.background_includes_bad_pixels)
    data["ninvfg"] = reflections.get_flags(flags.foreground_includes_bad_pixels)
    data["fbgd"] = reflections.get_flags(flags.failed_during_background_modelling)
    data["fsum"] = reflections.get_flags(flags.failed_during_summation)
    data["fprf"] = reflections.get_flags(flags.failed_during_profile_fitting)

    # Try to calculate the i over sigma for summation
    data["intensity.sum.ios"] = flex_ios(
        data["intensity.sum.value"], data["intensity.sum.variance"]
    )

    # Try to calculate the i over sigma for profile fitting
    try:
        data["intensity.prf.ios"] = flex_ios(
            data["intensity.prf.value"], data["intensity.prf.variance"]
        )
    except Exception:
        pass

    # Try to calculate the rmsd between observation and prediction
    try:
        xcal, ycal, zcal = data["xyzcal.px"].parts()
        xobs, yobs, zobs = data["xyzobs.px.value"].parts()
        data["xyz.rmsd"] = flex.sqrt(flex.pow2(xcal - xobs) + flex.pow2(ycal - yobs))
    except Exception:
        pass

    # Create the resolution binner
    resolution_binner = Binner(
        resolution_bins(experiment, data["miller_index"], n_resolution_bins)
    )

    # Create the frame binner object
    try:
        array_range = experiment.imageset.get_array_range()
    except Exception:
        array_range = (0, len(experiment.imageset))
    frame_binner = Binner(
        flex.int(range(array_range[0], array_range[1] + 1)).as_double()
    )

    # Create the overall report
    overall = overall_report(data)

    # Create high/low resolution reports
    hl_binner = resolution_binner.indexer(data["d"])
    high_summary = overall_report(select(data, hl_binner.indices(0)))
    low_summary = overall_report(select(data, hl_binner.indices(n_resolution_bins - 1)))
    high_summary["dmin"] = resolution_binner.bins()[0]
    high_summary["dmax"] = resolution_binner.bins()[1]
    low_summary["dmin"] = resolution_binner.bins()[n_resolution_bins - 1]
    low_summary["dmax"] = resolution_binner.bins()[n_resolution_bins]
    overall["dmin"] = high_summary["dmin"]
    overall["dmax"] = low_summary["dmax"]

    # Create the overall report
    summary = {
        "overall": overall,
        "low": low_summary,
        "high": high_summary,
    }

    # Create a report binned by resolution
    resolution = binned_report(resolution_binner, data["d"], data)

    # Create the report binned by image
    image = binned_report(frame_binner, data["xyzcal.px"].parts()[2], data)

    # Return the report
    return {"summary": summary, "resolution": resolution, "image": image}


class IntegrationReport(Report):
    """
    A class to store the integration report
    """

    def __init__(self, experiments, reflections):
        """
        Create the integration report

        :param experiments: The experiment list
        :param reflections: The reflection table
        """
        # Initialise the report class
        super().__init__()

        # Split the tables by experiment id
        tables = reflections.split_by_experiment_id()
        assert len(tables) == len(experiments)

        # Initialise the dictionary
        report_list = []

        # Generate an integration report for each experiment
        for i, (expr, data) in enumerate(zip(experiments, tables)):
            report_list.append(generate_integration_report(expr, data))

        # Construct the per image table
        table = Table()
        table.name = "integration.image.summary"
        table.title = "Summary vs image number"
        table.cols.append(("id", "ID"))
        table.cols.append(("image", "Image"))
        table.cols.append(("n_full", "# full"))
        table.cols.append(("n_part", "# part"))
        table.cols.append(("n_over", "# over"))
        table.cols.append(("n_ice", "# ice"))
        table.cols.append(("n_sum", "# sum"))
        table.cols.append(("n_prf", "# prf"))
        table.cols.append(("ibg", "Ibg"))
        table.cols.append(("ios_sum", "I/sigI\n (sum)"))
        table.cols.append(("ios_prf", "I/sigI\n (prf)"))
        table.cols.append(("cc_prf", "CC prf"))
        table.cols.append(("rmsd_xy", "RMSD XY"))
        for j, report in enumerate(report_list):
            report = report["image"]
            for i in range(len(report["bins"]) - 1):
                table.rows.append(
                    [
                        "%d" % j,
                        "%d" % (report["bins"][i] + 1),
                        "%d" % report["n_full"][i],
                        "%d" % report["n_partial"][i],
                        "%d" % report["n_overload"][i],
                        "%d" % report["n_ice"][i],
                        "%d" % report["n_summed"][i],
                        "%d" % report["n_fitted"][i],
                        f"{report['mean_background'][i]:.2f}",
                        f"{report['ios_sum'][i]:.2f}",
                        f"{report['ios_prf'][i]:.2f}",
                        f"{report['cc_prf'][i]:.2f}",
                        f"{report['rmsd_xy'][i]:.2f}",
                    ]
                )
        self.add_table(table)

        # Construct the per resolution table
        table = Table()
        table.name = "integration.resolution.summary"
        table.title = "Summary vs resolution"
        table.cols.append(("id", "ID"))
        table.cols.append(("dmin", "d min"))
        table.cols.append(("n_full", "# full"))
        table.cols.append(("n_part", "# part"))
        table.cols.append(("n_over", "# over"))
        table.cols.append(("n_ice", "# ice"))
        table.cols.append(("n_sum", "# sum"))
        table.cols.append(("n_prf", "# prf"))
        table.cols.append(("ibg", "Ibg"))
        table.cols.append(("ios_sum", "I/sigI\n (sum)"))
        table.cols.append(("ios_prf", "I/sigI\n (prf)"))
        table.cols.append(("cc_prf", "CC prf"))
        table.cols.append(("rmsd_xy", "RMSD XY"))
        for j, report in enumerate(report_list):
            report = report["resolution"]
            for i in range(len(report["bins"]) - 1):
                table.rows.append(
                    [
                        "%d" % j,
                        f"{report['bins'][i]:.2f}",
                        "%d" % report["n_full"][i],
                        "%d" % report["n_partial"][i],
                        "%d" % report["n_overload"][i],
                        "%d" % report["n_ice"][i],
                        "%d" % report["n_summed"][i],
                        "%d" % report["n_fitted"][i],
                        f"{report['mean_background'][i]:.2f}",
                        f"{report['ios_sum'][i]:.2f}",
                        f"{report['ios_prf'][i]:.2f}",
                        f"{report['cc_prf'][i]:.2f}",
                        f"{report['rmsd_xy'][i]:.2f}",
                    ]
                )
        self.add_table(table)

        # Create the overall table
        for j, report in enumerate(report_list):
            report = report["summary"]
            summary = report["overall"]
            high = report["high"]
            low = report["low"]

            table = Table()
            table.name = "integration.overall.summary"
            table.title = "Summary for experiment %d" % j
            table.cols.append(("item", "Item"))
            table.cols.append(("overall", "Overall"))
            table.cols.append(("low", "Low"))
            table.cols.append(("high", "High"))
            desc_fmt_key = [
                ("dmin", "%.2f", "dmin"),
                ("dmax", "%.2f", "dmax"),
                ("number fully recorded", "%d", "n_full"),
                ("number partially recorded", "%d", "n_partial"),
                ("number with invalid background pixels", "%d", "n_invalid_bg"),
                ("number with invalid foreground pixels", "%d", "n_invalid_fg"),
                ("number with overloaded pixels", "%d", "n_overload"),
                ("number in powder rings", "%d", "n_ice"),
                ("number processed with summation", "%d", "n_summed"),
                ("number processed with profile fitting", "%d", "n_fitted"),
                ("number failed in background modelling", "%d", "n_failed_background"),
                ("number failed in summation", "%d", "n_failed_summation"),
                ("number failed in profile fitting", "%d", "n_failed_fitting"),
                ("ibg", "%.2f", "mean_background"),
                ("i/sigi (summation)", "%.2f", "ios_sum"),
                ("i/sigi (profile fitting)", "%.2f", "ios_prf"),
                ("cc prf", "%.2f", "cc_prf"),
                ("cc_pearson sum/prf", "%.2f", "cc_pearson_sum_prf"),
                ("cc_spearman sum/prf", "%.2f", "cc_spearman_sum_prf"),
            ]
            for desc, fmt, key in desc_fmt_key:
                table.rows.append(
                    [desc, fmt % summary[key], fmt % low[key], fmt % high[key]]
                )
            self.add_table(table)


class ProfileModelReport(Report):
    """
    A class to store the profile model report
    """

    def __init__(self, experiments, fitter, reflections):
        """
        Create the integration report

        :param experiments: The experiment list
        :param profile_model: The profile model
        :param reflections: The reflection table
        """
        # Initialise the report class
        super().__init__()

        # Create the table
        table = Table()

        # Set the title
        table.name = "profile.summary"
        table.title = "Summary of profile model"

        # Add the columns
        table.cols.append(("id", "ID"))
        table.cols.append(("profile", "Profile"))
        table.cols.append(("created", "Created"))
        table.cols.append(("x", "X (px)"))
        table.cols.append(("y", "Y (px)"))
        table.cols.append(("z", "Z (im)"))
        table.cols.append(("n_reflections", "# reflections"))

        # Create the summary for each profile model
        for i in range(len(fitter)):
            model = fitter[i]
            for j in range(len(model)):
                table.rows.append(
                    [
                        "%d" % i,
                        "%d" % j,
                        f"{model.valid(j)}",
                        f"{model.coord(j)[0]:.2f}",
                        f"{model.coord(j)[1]:.2f}",
                        f"{model.coord(j)[2]:.2f}",
                        "%d" % model.n_reflections(j),
                    ]
                )

        # Add the table
        self.add_table(table)

        # Add the profiles
        for i in range(len(fitter)):
            model = fitter[i]
            for j in range(len(model)):
                if model.valid(j):
                    array = Array()
                    array.name = "profile.model.%d.%d" % (i, j)
                    array.title = "Profile model (id: %d, profile: %d)" % (i, j)
                    array.data = model.data(j)
                    self.add_array(array)


class ProfileValidationReport(Report):
    """
    A class to store the profile validation report
    """

    def __init__(self, experiments, profile_fitter, reflections, num_folds):
        """
        Create the integration report

        :param experiments: The experiment list
        :param profile_model: The profile model
        :param reflections: The reflection table
        """
        # Initialise the report class
        super().__init__()

        # Create the table
        table = Table()

        # Set the title
        table.name = "validation.summary"
        table.title = "Summary of profile validation "

        # Add the columns
        table.cols.append(("id", "ID"))
        table.cols.append(("subsample", "Sub-sample"))
        table.cols.append(("n_valid", "# validated"))
        table.cols.append(("cc", "<CC>"))
        table.cols.append(("nrmsd", "<NRMSD>"))

        # Split the reflections
        reflection_tables = reflections.split_by_experiment_id()
        assert len(reflection_tables) == len(experiments)
        assert len(profile_fitter) == num_folds

        # Create the summary for each profile model
        for i in range(len(reflection_tables)):
            reflection_table = reflection_tables[i]
            reflection_table = reflection_table.select(
                reflection_table.get_flags(reflection_table.flags.integrated_prf)
            )
            index = reflection_table["profile.index"]
            cc = reflection_table["profile.correlation"]
            nrmsd = reflection_table["profile.rmsd"]
            for j in range(num_folds):
                mask = index == j
                num_validated = mask.count(True)
                if num_validated == 0:
                    mean_cc = 0
                    mean_nrmsd = 0
                else:
                    mean_cc = flex.mean(cc.select(mask))
                    mean_nrmsd = flex.mean(nrmsd.select(mask))
                table.rows.append(
                    [
                        "%d" % i,
                        "%d" % j,
                        "%d" % num_validated,
                        f"{mean_cc:.2f}",
                        f"{mean_nrmsd:.2f}",
                    ]
                )

        # Add the table
        self.add_table(table)



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


def refine(params, reflections, experiments):
    if params.refinement.parameterisation.scan_varying:
        logger.warning(
            "scan_varying=True not supported in indexing: setting scan_varying=False"
        )
        params.refinement.parameterisation.scan_varying = False

    from dials.algorithms.refinement import RefinerFactory

    refiner = RefinerFactory.from_parameters_data_experiments(
        params, reflections, experiments
    )

    outliers = None
    refined = refiner.run()
    return refiner, refined, outliers



from __future__ import annotations

from unittest import mock

import pytest

from dxtbx.serialize import load

from dials.algorithms.scaling.observers import (
    ScalingHTMLContextManager,
    make_error_model_plots,
    make_filtering_plots,
    make_merging_stats_plots,
    make_outlier_plots,
    make_scaling_model_plots,
    print_scaling_model_error_summary,
    print_scaling_summary,
)
from dials.algorithms.scaling.scaling_library import scaled_data_as_miller_array
from dials.array_family import flex


@pytest.fixture
def test_data(dials_data):
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refl = location / "scaled_20_25.refl"
    expt = location / "scaled_20_25.expt"

    refls = flex.reflection_table.from_file(refl).split_by_experiment_id()
    for r in refls:
        r["s1c"] = r["s1"]  # just give any values for test
        r["s0c"] = r["s1"]  # just give any values for test
        r.set_flags(flex.bool(r.size(), True), r.flags.scaled)
    expts = load.experiment_list(expt, check_format=False)
    params = mock.Mock()
    params.physical.decay_restraint = 1e5
    for expt, refl in zip(expts, refls):
        expt.scaling_model.configure_components(refl, expt, params)
        expt.scaling_model.components["absorption"].update_reflection_data()

    return refls, expts, params


@pytest.fixture
def test_script(test_data):
    script = mock.Mock()
    refls, expts, params = test_data
    script.reflections = refls
    script.experiments = expts
    script.params = params
    script.scaled_miller_array = scaled_data_as_miller_array(refls, expts)
    script.merging_statistics_result = None
    script.anom_merging_statistics_result = None
    script.filtering_results = None
    return script


# test things which require refls, expts, params
def test_make_scaling_model_plots(test_data):
    refls, expts, __ = test_data
    graphs = make_scaling_model_plots(expts, refls)
    assert graphs
    assert len(graphs["scaling_model"]) == 12  # 2 models, 3 components each,
    # (4 plots for each absorption correction)


def test_print_scaling_model_error_summary(test_data):
    _, expts, __ = test_data
    msg = print_scaling_model_error_summary(expts)
    assert (
        msg
        == """Warning: Over half (83.67%) of model parameters have significant
uncertainty (sigma/abs(parameter) > 0.5), which could indicate a
poorly-determined scaling problem or overparameterisation.
"""
    )


def test_make_outlier_plots(test_data):
    refls, expts, _ = test_data
    graphs = make_outlier_plots(refls, expts)
    assert graphs
    assert len(graphs["outlier_plots"]) == 4


def test_make_error_model_plots(test_data):
    _, expts, params = test_data
    graphs = make_error_model_plots(params, expts)
    assert graphs
    assert "error_model_plots" in graphs


# test things which require a script object
def test_make_filtering_plots(test_script):
    plots = make_filtering_plots(test_script)
    assert plots


def test_make_merging_stats_plots(test_script):
    plots = make_merging_stats_plots(test_script)
    assert plots
    assert plots["image_range_tables"]


def test_print_scaling_summary(test_script):
    print_scaling_summary(test_script)


def test_ScalingHTMLContextManager(test_script, tmp_path):
    script = test_script
    script.params.output.html = str(tmp_path / "test.html")
    script.params.output.json = str(tmp_path / "test.json")
    with ScalingHTMLContextManager(script):
        pass

    assert (tmp_path / "test.html").is_file()
    assert (tmp_path / "test.json").is_file()



"""Test components of the dials.space_group program."""

from __future__ import annotations

import pytest

from dials.algorithms.symmetry.absences.laue_groups_info import (
    laue_groups,
    score_screw_axes,
    score_space_groups,
)
from dials.array_family import flex


def test_score_screw_axes_equivalent_axes():
    """Test the score_screw_axes function (when equivalent axes present)."""
    laue_group_info = laue_groups["P m -3"]

    # unique axis 21a, but should pick up 21b, 21c
    reflections = flex.reflection_table()
    reflections["miller_index"] = flex.miller_index(
        [(0, 1, 0), (0, 2, 0), (0, 0, 1), (0, 0, 2)]
    )
    reflections["intensity"] = flex.double([0.05, 100.0, 0.02, 100.0])
    reflections["variance"] = flex.double([1.0, 1.0, 1.0, 1.0])

    axes, scores = score_screw_axes(laue_group_info, reflections)
    assert len(scores) == 1
    assert len(axes) == 1
    assert axes[0].name == "21a"
    assert scores[0] > 0.99


def test_score_space_group():
    """Test scoring of space groups by combining axis scores."""

    # simple case, P 1 21 2, single score
    laue_group = laue_groups["P 1 2/m 1"]
    axis_scores = [0.98]  # 21 > P41212
    space_groups, scores = score_space_groups(axis_scores, laue_group)
    for sg, score in zip(space_groups, scores):
        if sg == "P 21":
            assert score == pytest.approx(0.98)
        elif sg == "P 2":
            assert score == pytest.approx(0.02)

    # More complex case - only use 42 score in space groups where not 41
    laue_group = laue_groups["P 4/m m m"]
    axis_scores = [0.95, 1.0, 0.95]  # 41, 21 and 42 score high > P41212
    space_groups, scores = score_space_groups(axis_scores, laue_group)
    for sg, score in zip(space_groups, scores):
        if sg == "P 41 21 2":
            assert score == pytest.approx(0.95)
        elif sg == "P 42 21 2":
            assert score == pytest.approx(0.05 * 0.95)
        elif sg == "P 4 21 2":
            assert score == pytest.approx(0.05 * 0.05)
        else:
            assert score == pytest.approx(0.0)

    # More complex case - only use 62, 63 score in space groups where not 61
    laue_group = laue_groups["P 6/m"]
    axis_scores = [0.95, 0.9, 0.85]  # 61, 62, 63
    space_groups, scores = score_space_groups(axis_scores, laue_group)
    for sg, score in zip(space_groups, scores):
        if sg == "P 61":
            assert score == pytest.approx(0.95)
        elif sg == "P 62":
            assert score == pytest.approx(0.05 * 0.9)
        elif sg == "P 63":
            assert score == pytest.approx(0.05 * 0.85)
        elif sg == "P 6":
            assert score == pytest.approx(0.05 * 0.1 * 0.15)



from __future__ import annotations

import pytest


def generate_image(xsize, ysize):
    from scitbx.array_family import flex

    image = flex.random_double(xsize * ysize)
    image.reshape(flex.grid(ysize, xsize))
    return image


def generate_mask(xsize, ysize):
    from scitbx.array_family import flex

    mask = flex.random_bool(xsize * ysize, 0.9)
    mask.reshape(flex.grid(ysize, xsize))
    return mask


def test_filter():
    from numpy import median

    from dials.algorithms.image.filter import median_filter

    xsize = 200
    ysize = 300
    kernel = (3, 3)

    image = generate_image(xsize, ysize)

    result = median_filter(image, kernel)

    eps = 1e-7

    for j in range(kernel[0], ysize - kernel[0]):
        for i in range(kernel[1], xsize - kernel[1]):
            j0 = j - kernel[0]
            j1 = j + kernel[0] + 1
            i0 = i - kernel[1]
            i1 = i + kernel[1] + 1
            pixels = image[j0:j1, i0:i1]
            value = median(pixels.as_numpy_array())
            assert result[j, i] == pytest.approx(value, abs=eps)


def test_masked_filter():
    from numpy import median

    from dials.algorithms.image.filter import median_filter

    xsize = 200
    ysize = 300
    kernel = (3, 3)

    image = generate_image(xsize, ysize)
    mask = generate_mask(xsize, ysize)
    result = median_filter(image, mask, kernel)
    eps = 1e-7

    for j in range(kernel[0], ysize - kernel[0]):
        for i in range(kernel[1], xsize - kernel[1]):
            if mask[j, i]:
                j0 = j - kernel[0]
                j1 = j + kernel[0] + 1
                i0 = i - kernel[1]
                i1 = i + kernel[1] + 1
                pixels = image[j0:j1, i0:i1]
                pmask = mask[j0:j1, i0:i1]
                pixels = pixels.as_1d().select(pmask.as_1d())
                if len(pixels) & 1:
                    value = median(pixels.as_numpy_array())
                else:
                    pixels = sorted(pixels)
                    value = pixels[len(pixels) // 2]
                assert result[j, i] == pytest.approx(value, abs=eps)



"""
Module of functions for handling operations on lists of reflection tables
and experiment lists.
"""


from __future__ import annotations

import copy
import logging

from orderedset import OrderedSet

from dxtbx.util import ersatz_uuid4

from dials.array_family import flex

logger = logging.getLogger("dials")

import iotbx.phil

phil_scope = iotbx.phil.parse(
    """
  dataset_selection {
    use_datasets = None
      .type = ints
      .help = "Choose a subset of datasets, based on the dataset id (as defined
               in the reflection table), to use from a multi-dataset input."
      .expert_level = 2
    exclude_datasets = None
      .type = ints
      .help = "Choose a subset of datasets, based on the dataset id (as defined
               in the reflection table), to exclude from a multi-dataset input."
      .expert_level = 2
  }
"""
)


def generate_experiment_identifiers(experiments, identifier_type="uuid"):
    """Generate unique identifiers for each experiment."""
    if identifier_type == "uuid":
        for expt in experiments:
            expt.identifier = ersatz_uuid4()
    elif identifier_type == "timestamp":
        pass


def sort_tables_to_experiments_order(reflection_tables, experiments):
    """If experiment identifiers are set, sort the order of reflection tables
    to match the order of the experiments.

    example for several single datasets in order
    input [r1("0"), r2("1")], [exp("0"), exp("1")]
    returns [r1("0"), r2("1")]

    example for several single datasets out of order
    input [r1("1"), r2("0")], [exp("0"), exp("1")]
    returns [r2("0"), r1("1")]

    example including a multi-dataset reflection table
    (e.g. datasets from command line: d1&2.refl d0.refl d0.expt d1&2.expt)
    input [r1("1", "2"), r2("0")], [exp("0"), exp("1"), exp("2")]
    returns [r2("0"), r1("1", "2")]

    Args:
        reflection_tables (list): A list of reflection tables (may contain multiple
            datasets per table)
        experiments: An ExperimentList

    Returns:
        (list): A list of sorted reflection tables to match the experiments order.
    """
    identifiers_list = []
    identifiers_by_table_idx = {}
    exp_id_to_table_idx = {}
    for i, table in enumerate(reflection_tables):
        id_values = table.experiment_identifiers().values()
        if id_values:
            identifiers_list.extend(id_values)
            identifiers_by_table_idx[i] = id_values
            for id_ in id_values:
                exp_id_to_table_idx[id_] = i

    expt_identiers = list(experiments.identifiers())
    if len(identifiers_list) == len(experiments) and set(identifiers_list) == set(
        expt_identiers
    ):  # all set so can now do the rearrangement
        sorted_tables = []
        for id_ in expt_identiers:
            # check if still need to add a table for this id
            if id_ in exp_id_to_table_idx:
                table_idx = exp_id_to_table_idx[id_]  # find table index
                sorted_tables.append(reflection_tables[table_idx])
                # now delete remaining ids from id>table map so that don't add twice
                for exp_id in identifiers_by_table_idx[table_idx]:
                    del exp_id_to_table_idx[exp_id]
        assert len(sorted_tables) == len(reflection_tables)
        return sorted_tables
    # else, identifiers not fully set, just return the tables.
    return reflection_tables


def renumber_table_id_columns(reflection_tables):
    """Renumber the id columns in the tables from 0..n-1
    If set, the experiment identifiers mapping is updated.

    Args:
        reflection_tables (list): A list of reflection tables

    Returns:
        (list): A list of reflection tables.
    """

    new_id_ = 0
    for table in reflection_tables:
        table_id_values = sorted(set(table["id"]).difference({-1}), reverse=True)
        highest_new_id = new_id_ + len(table_id_values) - 1
        expt_ids_dict = table.experiment_identifiers()
        new_ids_dict = {}
        new_id_ = highest_new_id
        orig_id = copy.deepcopy(table["id"])
        for val in table_id_values:
            sel = orig_id == val
            if val in expt_ids_dict:
                # only delete here, add new at end to avoid clashes of new/old ids
                new_ids_dict[new_id_] = expt_ids_dict[val]
                del expt_ids_dict[val]
            table["id"].set_selected(sel.iselection(), new_id_)
            new_id_ -= 1
        new_id_ = highest_new_id + 1
        if new_ids_dict:
            for i, v in new_ids_dict.items():
                expt_ids_dict[i] = v
    return reflection_tables


def parse_multiple_datasets(reflections):
    """
    Split a list of multi-dataset reflection tables, selecting on id

    If duplicate id values are found, the id columns are renumbered from 0..n-1,
    taking care of experiment identifiers if these are set.

    Args:
        reflections (list): a list of reflection tables, each of which may contain
            multiple datasets

    Returns:
        (list): a list of reflection tables corresponding to single datasets
    """
    single_reflection_tables = []
    dataset_id_list = []
    for refl_table in reflections:
        dataset_ids = set(refl_table["id"]).difference({-1})
        dataset_id_list.extend(list(dataset_ids))
        if len(dataset_ids) > 1:
            logger.info(
                "Detected existence of a multi-dataset reflection table \n"
                "containing %s datasets. \n",
                len(dataset_ids),
            )
            # FIXME fix split_by_experiment_id so that don't need to filter
            # unindxeded reflections here to get rid of id = -1
            if -1 in refl_table["id"]:
                refl_table = refl_table.select(refl_table["id"] != -1)
            result = refl_table.split_by_experiment_id()
            single_reflection_tables.extend(result)
        else:
            single_reflection_tables.append(refl_table)
    if len(dataset_id_list) != len(set(dataset_id_list)):  # need to reset some ids
        logger.warning(
            "Duplicate dataset ids found in different reflection tables. "
            "These will be treated as coming from separate datasets, and "
            "new dataset ids will be assigned for the whole dataset."
        )
        for new_id, (r, old_id) in enumerate(
            zip(single_reflection_tables, dataset_id_list)
        ):
            r["id"] = flex.int(r.size(), new_id)
            if list(r.experiment_identifiers()):  # if identifiers, need to update
                expid = r.experiment_identifiers()[old_id]
                del r.experiment_identifiers()[old_id]
                r.experiment_identifiers()[new_id] = expid
    return single_reflection_tables


def assign_unique_identifiers(experiments, reflections, identifiers=None):
    """
    Assign unique experiment identifiers to experiments and reflections lists.

    If experiment identifiers are not set for some datasets, then new unique
    identifiers are given to those, and the 'id' column for all reflection tables
    are set sequentially from 0..n-1.

    Args:
        experiments: An ExperimentList
        reflections (list): A list of reflection tables

    Returns:
        (tuple): tuple containing:
            experiments: The updated ExperimentList
            reflections (list): A list of the updated reflection tables

    Raises:
        ValueError: If the number of reflection tables and experiments are
            unequal (and identifiers if specified). Also raised if the existing
            identifiers are corrupted.
    """
    if len(experiments) != len(reflections):
        raise ValueError(
            "The experiments and reflections lists are unequal in length: %s & %s"
            % (len(experiments), len(reflections))
        )
    # if identifiers given, use these to set the identifiers
    if identifiers:
        if len(identifiers) != len(reflections):
            raise ValueError(
                "The identifiers and reflections lists are unequal in length: %s & %s"
                % (len(identifiers), len(reflections))
            )
        for i, (exp, refl) in enumerate(zip(experiments, reflections)):
            exp.identifier = identifiers[i]
            for k in refl.experiment_identifiers().keys():
                del refl.experiment_identifiers()[k]
            refl.experiment_identifiers()[i] = identifiers[i]
            refl["id"] = flex.int(refl.size(), i)
    # Validate the existing identifiers, or the ones just set
    used_str_ids = []
    for exp, refl in zip(experiments, reflections):
        if exp.identifier != "":
            if list(refl.experiment_identifiers().values()) != [exp.identifier]:
                raise ValueError(
                    "Corrupted identifiers, please check input: in reflections: %s, in experiment: %s"
                    % (list(refl.experiment_identifiers().values()), exp.identifier)
                )
            used_str_ids.append(exp.identifier)

    if len(set(used_str_ids)) != len(reflections):
        # if not all set, then need to fill in the rest. Keep the identifier if
        # it is already set, and reset table id column from 0..n-1
        for i, (exp, refl) in enumerate(zip(experiments, reflections)):
            if exp.identifier == "":
                strid = ersatz_uuid4()
                exp.identifier = strid
                refl.experiment_identifiers()[i] = strid
            else:
                k = list(refl.experiment_identifiers().keys())[0]
                expid = list(refl.experiment_identifiers().values())[0]
                del refl.experiment_identifiers()[k]
                refl.experiment_identifiers()[i] = expid
            refl["id"] = flex.int(refl.size(), i)
    return experiments, reflections


def select_datasets_on_ids(
    experiments, reflection_table_list, exclude_datasets=None, use_datasets=None
):
    # transform dataset ids to identifiers
    if not use_datasets and not exclude_datasets:
        return experiments, reflection_table_list
    if use_datasets and exclude_datasets:
        raise ValueError(
            "The options use_datasets and exclude_datasets cannot be used in conjunction."
        )

    id_map = {}
    for table in reflection_table_list:
        id_map.update(table.experiment_identifiers())
    if exclude_datasets:
        identifiers_to_exclude = []
        for k in exclude_datasets:
            if int(k) not in id_map:
                raise ValueError(
                    """Attempting to select datasets based on identifiers that
are not found in the experiment list / reflection tables."""
                )
            identifiers_to_exclude.append(id_map[int(k)])
        return select_datasets_on_identifiers(
            experiments, reflection_table_list, exclude_datasets=identifiers_to_exclude
        )
    elif use_datasets:
        identifiers_to_use = []
        for k in use_datasets:
            if int(k) not in id_map:
                raise ValueError(
                    """Attempting to select datasets based on identifiers that
are not found in the experiment list / reflection tables."""
                )
            identifiers_to_use.append(id_map[int(k)])
        return select_datasets_on_identifiers(
            experiments, reflection_table_list, use_datasets=identifiers_to_use
        )


def select_datasets_on_identifiers(
    experiments, reflection_table_list, exclude_datasets=None, use_datasets=None
):
    """
    Select a subset of experiments and reflection tables based on identifiers.

    This performs a similar function to the select/remove_on_experiment_identifiers
    methods of ExperimentList and reflection_table, with additional logic to handle
    the case of a list of reflection tables, rather than a single one and to catch
    bad input. Does not require reflection tables containing data from multiple
    experiments to be split.

    Args:
        experiments: An ExperimentList
        reflection_table_list (list): a list of reflection tables
        exclude_datasets (list): a list of experiment_identifiers to exclude
        use_datasets (list): a list of experiment_identifiers to use

    Returns:
        (tuple): tuple containing:
            experiments: The updated ExperimentList
            list_of_reflections (list): A list of the updated reflection tables

    Raises:
        ValueError: If both use_datasets and exclude datasets are used, if not all
            experiment identifiers are set, if an identifier in exclude_datasets
            or use_datasets is not in the list.
    """
    if not use_datasets and not exclude_datasets:
        return experiments, reflection_table_list
    if use_datasets and exclude_datasets:
        raise ValueError(
            "The options use_datasets and exclude_datasets cannot be used in conjunction."
        )
    if experiments.identifiers().count("") > 0:
        raise ValueError(
            f"""
            Not all experiment identifiers set in the ExperimentList.
            Current identifiers set as: {list(experiments.identifiers())}"""
        )
    list_of_reflections = []
    if use_datasets:
        total_found = 0
        for reflection_table in reflection_table_list:
            expids = list(reflection_table.experiment_identifiers().values())
            expids_in_this_table = list(set(use_datasets).intersection(set(expids)))
            total_found += len(expids_in_this_table)
            if expids_in_this_table:  # if none, then no datasets wanted from table
                list_of_reflections.append(
                    reflection_table.select_on_experiment_identifiers(
                        expids_in_this_table
                    )
                )
        if total_found != len(use_datasets):
            raise ValueError(
                """Attempting to select datasets based on identifiers that
are not found in the experiment list / reflection tables."""
            )
        experiments.select_on_experiment_identifiers(use_datasets)
    elif exclude_datasets:
        if len(
            set(exclude_datasets).intersection(set(experiments.identifiers()))
        ) != len(set(exclude_datasets)):
            raise ValueError(
                """Attempting to exclude datasets based on identifiers that
are not found in the experiment list / reflection tables."""
            )
        for reflection_table in reflection_table_list:
            expids = list(reflection_table.experiment_identifiers().values())
            expids_in_this_table = list(set(exclude_datasets).intersection(set(expids)))
            if expids_in_this_table:  # only append if data left after removing
                r_table = reflection_table.remove_on_experiment_identifiers(
                    expids_in_this_table
                )
                if r_table.size() > 0:
                    list_of_reflections.append(r_table)
            else:
                list_of_reflections.append(reflection_table)
        experiments.remove_on_experiment_identifiers(exclude_datasets)
    return experiments, list_of_reflections


def update_imageset_ids(experiments, reflections):
    """For a list of input experiments and reflections (each containing one
    sweep), update or add the imageset_id column to the data to match the order
    in the experiment list.

    This means that when the reflection tables are combined, the data is correct.
    """
    # input a list of ordered matching experiments and reflection tables.

    next_iset_id = 0
    imagesets_found = OrderedSet()
    for expt, table in zip(experiments, reflections):
        if "imageset_id" in table:
            assert len(set(table["imageset_id"])) == 1
        iset = expt.imageset
        if iset not in imagesets_found:
            imagesets_found.add(iset)
            table["imageset_id"] = flex.int(table.size(), next_iset_id)
            next_iset_id += 1
        else:
            iset_id = imagesets_found.index(iset)
            table["imageset_id"] = flex.int(table.size(), iset_id)
    return reflections



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1
# DIALS_ENABLE_COMMAND_LINE_COMPLETION

from __future__ import annotations

import logging
import math

import libtbx.phil
from scitbx import matrix
from scitbx.array_family import flex

import dials.util
from dials.algorithms.indexing.indexer import find_max_cell
from dials.command_line.search_beam_position import run_dps
from dials.util.reciprocal_lattice import Render3d

try:
    import matplotlib
except ImportError:
    exit()  # To pass through the "make" step, for graphics-free HPC build

# Offline backend
matplotlib.use("Agg")

logger = logging.getLogger("dials.command_line.rl_png")

help_message = """
Generate a png of the strong spots from spotfinding in reciprocal space.

Examples::

  dials.rl_png imported.expt strong.refl

  dials.rl_png indexed.expt indexed.refl
"""

phil_scope = libtbx.phil.parse(
    """
include scope dials.util.reciprocal_lattice.phil_scope

marker_size = 5
  .type = int(value_min=1)

basis_vector_search {
  n_solutions = 3
    .type = int
}
plot {
  size_inches = 10,10
    .type = floats(size=2, value_min=0)
}
""",
    process_includes=True,
)


class ReciprocalLatticePng(Render3d):
    def __init__(self, settings=None):
        Render3d.__init__(self, settings=settings)
        self.viewer = PngScene(settings=self.settings)


class PngScene:
    def __init__(self, settings):
        self.settings = settings
        self.rotation_axis = None
        self.beam_vector = None
        self.points = None
        self.colors = None
        self.palette = None

    def set_rotation_axis(self, axis):
        self.rotation_axis = axis

    def set_beam_vector(self, beam):
        self.beam_vector = beam

    def set_points(self, points):
        self.points = points

    def set_points_data(self, reflections):
        # we do not label reciprocal lattice points here
        pass

    def set_colors(self, colors):
        # convert whites to black (background is white)
        colors.set_selected((colors.norms() == math.sqrt(3)), (0, 0, 0))
        self.colors = colors

    def set_palette(self, palette):
        self.palette = palette

    def set_reciprocal_lattice_vectors(self, *args, **kwargs):
        # we do not draw reciprocal lattice vectors at this time
        pass

    def set_reciprocal_crystal_vectors(self, *args, **kwargs):
        # we do not draw reciprocal crystal vectors at this time either
        pass

    def project_2d(self, n):
        d = self.points.dot(n.elems)
        p = d * flex.vec3_double(len(d), n.elems)

        points2d = self.points - p

        x = matrix.col((1, 0, 0))
        if x.angle(n) == 0 or x.angle(-n) == 0:
            x = matrix.col((0, 1, 0))

        x = (x - x.dot(n) * n).normalize()

        y = x.cross(n)
        # assert y.angle(x, deg=True) == 90
        # assert y.angle(matrix.col(n), deg=True) == 90

        px2d = points2d.dot(x)
        py2d = points2d.dot(y)

        return px2d, py2d

    def plot(self, filename, n=(1, 0, 0)):
        from matplotlib import pyplot

        n = matrix.col(n).normalize()
        x, y = self.project_2d(n)
        fig = pyplot.figure(figsize=self.settings.plot.size_inches)
        pyplot.scatter(
            x.as_numpy_array(),
            y.as_numpy_array(),
            marker="+",
            s=self.settings.marker_size,
            c=list(self.colors),
        )
        pyplot.title("Plane normal: (%.2g, %.2g, %.2g)" % (n.elems))
        fig.savefig(filename)
        pyplot.close()


@dials.util.show_mail_handle_errors()
def run(args=None):
    from dials.util import log
    from dials.util.options import (
        ArgumentParser,
        reflections_and_experiments_from_files,
    )

    usage = "dials.rl_png [options] experiments.json observations.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args)
    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if len(experiments) == 0 or len(reflections) == 0:
        parser.print_help()
        exit(0)

    # Configure the logging
    log.config(logfile="dials.rl_png.log")

    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil != "":
        logger.info("The following parameters have been modified:\n")
        logger.info(diff_phil)

    reflections = reflections[0]

    f = ReciprocalLatticePng(settings=params)
    f.load_models(experiments, reflections)

    rotation_axis = matrix.col(experiments[0].goniometer.get_rotation_axis())
    s0 = matrix.col(experiments[0].beam.get_s0())

    e1 = rotation_axis.normalize()
    e2 = s0.normalize()
    e3 = e1.cross(e2).normalize()

    f.viewer.plot("rl_rotation_axis.png", n=e1.elems)
    f.viewer.plot("rl_beam_vector", n=e2.elems)
    f.viewer.plot("rl_e3.png", n=e3.elems)

    n_solutions = params.basis_vector_search.n_solutions

    if experiments.crystals().count(None) < len(experiments):
        for i, c in enumerate(experiments.crystals()):
            A = matrix.sqr(c.get_A())

            direct_matrix = A.inverse()
            a = direct_matrix[:3]
            b = direct_matrix[3:6]
            c = direct_matrix[6:9]

            prefix = ""
            if len(experiments.crystals()) > 1:
                prefix = "%i_" % (i + 1)

            f.viewer.plot(f"rl_{prefix}a.png", n=a)
            f.viewer.plot(f"rl_{prefix}b.png", n=b)
            f.viewer.plot(f"rl_{prefix}c.png", n=c)

    elif n_solutions:
        if "imageset_id" not in reflections:
            reflections["imageset_id"] = reflections["id"]

        reflections.centroid_px_to_mm(experiments)

        reflections.map_centroids_to_reciprocal_space(experiments)

        if params.d_min is not None:
            d_spacings = 1 / reflections["rlp"].norms()
            sel = d_spacings > params.d_min
            reflections = reflections.select(sel)

        # derive a max_cell from mm spots
        max_cell = find_max_cell(
            reflections, max_cell_multiplier=1.3, step_size=45
        ).max_cell

        result = run_dps(experiments[0], reflections, max_cell)
        if result:
            solutions = [matrix.col(v) for v in result["solutions"]]
            for i in range(min(n_solutions, len(solutions))):
                v = solutions[i]
                f.viewer.plot(f"rl_solution_{i + 1}.png", n=v.elems)


if __name__ == "__main__":
    run()



from __future__ import annotations

import math
import os

import pytest

from scitbx import matrix


class RayPredictor:
    def __init__(self, dials_regression):
        import dxtbx
        from iotbx.xds import integrate_hkl, xparm
        from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter

        from dials.algorithms.spot_prediction import (
            IndexGenerator,
            ScanStaticRayPredictor,
        )
        from dials.util import ioutil

        # The XDS files to read from
        integrate_filename = os.path.join(
            dials_regression, "data", "sim_mx", "INTEGRATE.HKL"
        )
        gxparm_filename = os.path.join(dials_regression, "data", "sim_mx", "GXPARM.XDS")

        # Read the XDS files
        self.integrate_handle = integrate_hkl.reader()
        self.integrate_handle.read_file(integrate_filename)
        self.gxparm_handle = xparm.reader()
        self.gxparm_handle.read_file(gxparm_filename)

        # Get the parameters we need from the GXPARM file
        models = dxtbx.load(gxparm_filename)
        self.beam = models.get_beam()
        self.gonio = models.get_goniometer()
        self.detector = models.get_detector()
        self.scan = models.get_scan()

        # Get crystal parameters
        self.space_group_type = ioutil.get_space_group_type_from_xparm(
            self.gxparm_handle
        )
        cfc = coordinate_frame_converter(gxparm_filename)
        a_vec = cfc.get("real_space_a")
        b_vec = cfc.get("real_space_b")
        c_vec = cfc.get("real_space_c")
        self.unit_cell = cfc.get_unit_cell()
        self.ub_matrix = matrix.sqr(a_vec + b_vec + c_vec).inverse()

        # Get the minimum resolution in the integrate file
        d = [self.unit_cell.d(h) for h in self.integrate_handle.hkl]
        self.d_min = min(d)
        # extend the resolution shell by epsilon>0
        # to account for rounding artifacts on 32-bit platforms
        self.d_min = self.d_min - 1e-15

        # Get the number of frames from the max z value
        xcal, ycal, zcal = zip(*self.integrate_handle.xyzcal)
        self.scan.set_image_range(
            (
                self.scan.get_image_range()[0],
                self.scan.get_image_range()[0] + int(math.ceil(max(zcal))),
            )
        )

        # Print stuff
        #        print self.beam
        #        print self.gonio
        #        print self.detector
        #        print self.scan

        # Create the index generator
        self.generate_indices = IndexGenerator(
            self.unit_cell, self.space_group_type, self.d_min
        )

        s0 = self.beam.get_s0()
        m2 = self.gonio.get_rotation_axis()
        fixed_rotation = self.gonio.get_fixed_rotation()
        setting_rotation = self.gonio.get_setting_rotation()
        UB = self.ub_matrix
        dphi = self.scan.get_oscillation_range(deg=False)

        # Create the ray predictor
        self.predict_rays = ScanStaticRayPredictor(
            s0, m2, fixed_rotation, setting_rotation, dphi
        )

        # Predict the spot locations
        self.reflections = self.predict_rays(self.generate_indices.to_array(), UB)


@pytest.fixture(scope="session")
def raypredictor(dials_regression):
    return RayPredictor(dials_regression)


def test_miller_index_set(raypredictor):
    """Ensure we have the whole set of miller indices"""
    gen_hkl = {r["miller_index"] for r in raypredictor.reflections.rows()}
    missing = []
    for hkl in raypredictor.integrate_handle.hkl:
        if hkl not in gen_hkl:
            missing.append(hkl)
    assert len(missing) == 0, "%d out of %d reflections not in set, including %s" % (
        len(missing),
        len(raypredictor.integrate_handle.hkl),
        str(missing[0]),
    )


def test_rotation_angles(raypredictor):
    """Ensure the rotation angles agree with XDS"""

    # Create a dict of lists of xy for each hkl
    gen_phi = {}
    for r in raypredictor.reflections.rows():
        hkl = r["miller_index"]
        phi = r["phi"]
        try:
            a = gen_phi[hkl]
            a.append(phi)
            gen_phi[hkl] = a
        except KeyError:
            gen_phi[hkl] = [phi]

    # For each hkl in the xds file
    for hkl, xyz in zip(
        raypredictor.integrate_handle.hkl, raypredictor.integrate_handle.xyzcal
    ):

        # Calculate the XDS phi value
        xds_phi = (
            raypredictor.scan.get_oscillation(deg=False)[0]
            + xyz[2] * raypredictor.scan.get_oscillation(deg=False)[1]
        )

        # Select the nearest xy to use if there are 2
        my_phi = gen_phi[hkl]
        if len(my_phi) == 2:
            my_phi0 = my_phi[0]
            my_phi1 = my_phi[1]
            diff0 = abs(xds_phi - my_phi0)
            diff1 = abs(xds_phi - my_phi1)
            if diff0 < diff1:
                my_phi = my_phi0
            else:
                my_phi = my_phi1
        else:
            my_phi = my_phi[0]

        # Check the Phi values are the same
        assert xds_phi == pytest.approx(my_phi, abs=0.1)


def test_beam_vectors(raypredictor):
    """Ensure |s1| == |s0|"""
    s0_length = matrix.col(raypredictor.beam.get_s0()).length()
    for r in raypredictor.reflections.rows():
        s1 = r["s1"]
        s1_length = matrix.col(s1).length()
        assert s0_length == pytest.approx(s1_length, abs=1e-7)


def test_new(raypredictor):
    from dials.algorithms.spot_prediction import IndexGenerator, ScanStaticRayPredictor

    # Create the index generator
    raypredictor.generate_indices = IndexGenerator(
        raypredictor.unit_cell, raypredictor.space_group_type, raypredictor.d_min
    )

    s0 = raypredictor.beam.get_s0()
    m2 = raypredictor.gonio.get_rotation_axis()
    fixed_rotation = raypredictor.gonio.get_fixed_rotation()
    setting_rotation = raypredictor.gonio.get_setting_rotation()
    UB = raypredictor.ub_matrix
    dphi = raypredictor.scan.get_oscillation_range(deg=False)

    # Create the ray predictor
    raypredictor.predict_rays = ScanStaticRayPredictor(
        s0, m2, fixed_rotation, setting_rotation, dphi
    )

    # Predict the spot locations
    raypredictor.reflections2 = []
    for h in raypredictor.generate_indices.to_array():
        rays = raypredictor.predict_rays(h, UB)
        for ray in rays:
            if ray.angle >= dphi[0] and ray.angle <= dphi[1]:
                raypredictor.reflections2.append(ray)

    assert len(raypredictor.reflections) == len(raypredictor.reflections2)
    for r1, r2 in zip(raypredictor.reflections.rows(), raypredictor.reflections2):
        assert all(a == pytest.approx(b, abs=1e-7) for a, b in zip(r1["s1"], r2.s1))
        assert r1["phi"] == pytest.approx(r2.angle, abs=1e-7)
        assert r1["entering"] == r2.entering


def test_new_from_array(raypredictor):
    from dials.algorithms.spot_prediction import IndexGenerator, ScanStaticRayPredictor

    # Create the index generator
    raypredictor.generate_indices = IndexGenerator(
        raypredictor.unit_cell, raypredictor.space_group_type, raypredictor.d_min
    )

    s0 = raypredictor.beam.get_s0()
    m2 = raypredictor.gonio.get_rotation_axis()
    fixed_rotation = raypredictor.gonio.get_fixed_rotation()
    setting_rotation = raypredictor.gonio.get_setting_rotation()
    UB = raypredictor.ub_matrix
    dphi = raypredictor.scan.get_oscillation_range(deg=False)

    # Create the ray predictor
    raypredictor.predict_rays = ScanStaticRayPredictor(
        s0, m2, fixed_rotation, setting_rotation, dphi
    )

    # Predict the spot locations
    h = raypredictor.generate_indices.to_array()
    reflections = raypredictor.predict_rays(h, UB)
    raypredictor.reflections3 = []
    for r in reflections.rows():
        if r["phi"] >= dphi[0] and r["phi"] <= dphi[1]:
            raypredictor.reflections3.append(r)

    assert len(raypredictor.reflections) == len(raypredictor.reflections3)
    for r1, r2 in zip(raypredictor.reflections.rows(), raypredictor.reflections3):
        assert all(a == pytest.approx(b, abs=1e-7) for a, b in zip(r1["s1"], r2["s1"]))
        assert r1["phi"] == pytest.approx(r2["phi"], abs=1e-7)
        assert r1["entering"] == r2["entering"]


def test_scan_varying(raypredictor):
    import scitbx.math

    from dials.algorithms.spot_prediction import (
        ReekeIndexGenerator,
        ScanVaryingRayPredictor,
    )

    s0 = raypredictor.beam.get_s0()
    m2 = raypredictor.gonio.get_rotation_axis()

    # For quick comparison look at reflections on one frame only
    frame = 0
    angle_beg = raypredictor.scan.get_angle_from_array_index(frame, deg=False)
    angle_end = raypredictor.scan.get_angle_from_array_index(frame + 1, deg=False)
    frame0_refs = raypredictor.reflections.select(
        (raypredictor.reflections["phi"] >= angle_beg)
        & (raypredictor.reflections["phi"] <= angle_end)
    )

    # Get UB matrices at beginning and end of frame
    r_osc_beg = matrix.sqr(
        scitbx.math.r3_rotation_axis_and_angle_as_matrix(
            axis=m2, angle=angle_beg, deg=False
        )
    )
    UB_beg = r_osc_beg * raypredictor.ub_matrix
    r_osc_end = matrix.sqr(
        scitbx.math.r3_rotation_axis_and_angle_as_matrix(
            axis=m2, angle=angle_end, deg=False
        )
    )
    UB_end = r_osc_end * raypredictor.ub_matrix

    # Get indices
    r = ReekeIndexGenerator(
        UB_beg,
        UB_end,
        raypredictor.space_group_type,
        m2,
        s0,
        raypredictor.d_min,
        margin=1,
    )
    h = r.to_array()

    # Fn to loop through hkl applying a prediction function to each and testing
    # the results are the same as those from the ScanStaticRayPredictor
    def test_each_hkl(hkl_list, predict_fn):
        DEG2RAD = math.pi / 180.0
        count = 0
        for hkl in hkl_list:
            ray = predict_fn(hkl)
            if ray is not None:
                count += 1
                ref = frame0_refs.select(frame0_refs["miller_index"] == hkl)[0]
                assert ref["entering"] == ray.entering
                assert ref["phi"] == pytest.approx(
                    ray.angle * DEG2RAD, abs=1e-6
                )  # ray angle is in degrees (!)
                assert ref["s1"] == pytest.approx(ray.s1, abs=1e-6)
        # ensure all reflections were matched
        assert count == len(frame0_refs)

    # Create the ray predictor
    sv_predict_rays = ScanVaryingRayPredictor(
        s0,
        m2,
        raypredictor.scan.get_array_range()[0],
        raypredictor.scan.get_oscillation(),
        raypredictor.d_min,
    )

    # Test with the method that allows only differing UB matrices
    test_each_hkl(h, lambda x: sv_predict_rays(x, UB_beg, UB_end, frame))

    # Now repeat prediction using the overload that allows for different s0
    # at the beginning and end of the frame. Here, pass in the same s0 each time
    # and the result should be the same as before
    test_each_hkl(h, lambda x: sv_predict_rays(x, UB_beg, UB_end, s0, s0, frame))



from __future__ import annotations

import math
import random

from dxtbx.serialize import load
from scitbx import matrix

from dials.algorithms.profile_model.gaussian_rs import CoordinateSystem, transform


def test_run(dials_data):
    sequence = load.imageset(
        dials_data("centroid_test_data", pathlib=True) / "sweep.json"
    )

    # Get the models
    beam = sequence.get_beam()
    detector = sequence.get_detector()
    gonio = sequence.get_goniometer()
    assert len(detector) == 1

    # Get some stuff
    s0 = beam.get_s0()
    m2 = gonio.get_rotation_axis()
    image_size = detector[0].get_image_size()

    # Get a random s1/phi
    i = random.uniform(0, image_size[0])
    j = random.uniform(1, image_size[1])
    s1 = matrix.col(detector[0].get_pixel_lab_coord((i, j)))
    s1 = s1.normalize() * matrix.col(s0).length()
    phi = random.uniform(0, 5)
    x0 = int(math.floor(i - 10))
    y0 = int(math.floor(j - 10))

    # Set some parameters
    sigma_divergence = beam.get_sigma_divergence(deg=False)
    delta_divergence = 3 * sigma_divergence
    grid_half_size = 4
    step_size = (delta_divergence / grid_half_size, delta_divergence / grid_half_size)

    # Create the coordinate system
    cs = CoordinateSystem(m2, s0, s1, phi)

    # Create the map of s1 coordinates
    s1_map = transform.beam_vector_map(detector[0], beam, True)

    # Create the grid index generator
    generate_indices = transform.GridIndexGenerator(
        cs, x0, y0, step_size, grid_half_size, s1_map
    )

    for j in range(0, 20):
        for i in range(0, 20):
            xx = x0 + i
            yy = y0 + j
            if xx < 0 or yy < 0 or xx >= image_size[0] or yy >= image_size[0]:
                continue

            # Get the grid indices
            gi_1, gj_1 = generate_indices(j, i)

            # Get the grid indices
            xyz = matrix.col(detector[0].get_pixel_lab_coord((x0 + i, y0 + j)))
            xyz = xyz.normalize() * matrix.col(s0).length()
            c1, c2 = matrix.col(cs.from_beam_vector(xyz))
            gi_2 = grid_half_size + c1 / step_size[0] + 0.5
            gj_2 = grid_half_size + c2 / step_size[1] + 0.5

            # Check both are the same
            eps = 1e-7
            assert abs(gj_1 - gj_2) <= eps, (gi_1, gi_2, gj_1, gj_2)
            assert abs(gi_1 - gi_2) <= eps, (gi_1, gi_2, gj_1, gj_2)



from __future__ import annotations

import random

from dials.algorithms.background.simple import (
    NormalOutlierRejector,
    NSigmaOutlierRejector,
    TruncatedOutlierRejector,
)
from dials.algorithms.shoebox import MaskCode
from dials.algorithms.simulation.generate_test_reflections import (
    random_background_plane2,
)
from dials.array_family import flex


def generate_shoebox(size, mean, nforeground, ninvalid):
    data = flex.double(flex.grid(size), 0.0)
    mask = flex.int(flex.grid(size), MaskCode.Valid | MaskCode.Background)
    random_background_plane2(data, mean, 0, 0, 0)
    for i in random.sample(range(len(data)), ninvalid):
        mask[i] &= ~MaskCode.Valid
    for i in random.sample(range(len(data)), nforeground):
        mask[i] |= MaskCode.Foreground
        mask[i] &= ~MaskCode.Background
    return data, mask


def assert_basic_mask_is_correct(mask, ninvalid, nforeground):
    invalid = flex.size_t(i for i in range(len(mask)) if not mask[i] & MaskCode.Valid)
    foreground = flex.size_t(
        i for i in range(len(mask)) if mask[i] & MaskCode.Foreground
    )
    background = flex.size_t(
        i for i in range(len(mask)) if mask[i] & MaskCode.Background
    )
    background_used = flex.size_t(
        i for i in range(len(mask)) if mask[i] & MaskCode.BackgroundUsed
    )
    background_valid = flex.size_t(
        i
        for i in range(len(mask))
        if mask[i] & MaskCode.Valid and mask[i] & MaskCode.Background
    )
    assert len(invalid) == ninvalid
    assert len(foreground) == nforeground
    assert len(background) == len(mask) - len(foreground)
    assert len(background_used) < len(background)
    assert len(set(background).intersection(foreground)) == 0
    assert len(set(background).intersection(background_used)) == len(background_used)
    return invalid, foreground, background, background_used, background_valid


def test_truncated():
    lower = 0.01
    upper = 0.01
    reject = TruncatedOutlierRejector(lower, upper)
    size = (9, 9, 9)
    ninvalid = 5
    nforeground = 20
    mean = 20

    def assert_is_correct(data, mask):
        (
            invalid,
            foreground,
            background,
            background_used,
            background_valid,
        ) = assert_basic_mask_is_correct(mask, ninvalid, nforeground)
        ntot = len(background_valid)
        i0 = int(lower * ntot / 2.0)
        i1 = int(upper * ntot / 2.0)
        nexp = ntot - i1 - i0
        assert len(background_used) == nexp
        index_exp = sorted(mask.select(background_valid), key=lambda x: data[x])[i0:i1]
        assert all(ii == jj for ii, jj in zip(index_exp, background_used))

    for i in range(10):
        data, mask = generate_shoebox(size, mean, nforeground, ninvalid)
        reject(data, mask)
        assert_is_correct(data, mask)


def test_nsigma():
    lower = 3.0
    upper = 3.0
    reject = NSigmaOutlierRejector(lower, upper)
    size = (9, 9, 9)
    ninvalid = 5
    nforeground = 20
    mean = 20

    def assert_is_correct(data, mask):
        (
            invalid,
            foreground,
            background,
            background_used,
            background_valid,
        ) = assert_basic_mask_is_correct(mask, ninvalid, nforeground)
        subdata = data.select(background_valid)
        mv = flex.mean_and_variance(subdata)
        m = mv.mean()
        s = mv.unweighted_sample_standard_deviation()
        p0 = m - lower * s
        p1 = m + upper * s
        mask = (subdata >= p0) & (subdata <= p1)
        exp = background_valid.select(mask)
        assert len(exp) == len(background_used)
        assert all(ii == jj for ii, jj in zip(exp, background_used))

    for i in range(10):
        data, mask = generate_shoebox(size, mean, nforeground, ninvalid)
        reject(data, mask)
        assert_is_correct(data, mask)


def test_normal():
    min_data = 10
    reject = NormalOutlierRejector(min_data)
    size = (9, 9, 9)
    ninvalid = 5
    nforeground = 20
    mean = 20

    def assert_is_correct(data, mask):
        assert_basic_mask_is_correct(mask, ninvalid, nforeground)

    for i in range(10):
        data, mask = generate_shoebox(size, mean, nforeground, ninvalid)
        reject(data, mask)
        assert_is_correct(data, mask)



from __future__ import annotations

import math

import wx
from wx.lib.agw.floatspin import EVT_FLOATSPIN, FloatSpin

import cctbx.miller
from cctbx.crystal import symmetry
from scitbx.matrix import col


class UCSettingsFrame(wx.MiniFrame):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        szr = wx.BoxSizer(wx.VERTICAL)
        self.phil_params = args[0].params
        panel = UCSettingsPanel(self)
        self.SetSizer(szr)
        szr.Add(panel, 1, wx.EXPAND)
        szr.Fit(panel)
        self.panel = panel
        self.sizer = szr
        self.Fit()
        self.Bind(wx.EVT_CLOSE, lambda evt: self.Destroy(), self)


class UCSettingsPanel(wx.Panel):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

        self.phil_params = args[0].phil_params

        # Needed to draw and delete the rings.  XXX Applies to
        # calibration_frame as well?
        self._pyslip = self.GetParent().GetParent().pyslip

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        # Number of decimal digits for distances.
        self.digits = 2

        # Wavelength control.
        beam = self._pyslip.tiles.raw_image.get_beam()
        self._wavelength = beam.get_wavelength()

        # Unit cell controls.
        if self.phil_params.calibrate_unit_cell.unit_cell is not None:
            self._cell = list(
                self.phil_params.calibrate_unit_cell.unit_cell.parameters()
            )
        else:
            self._cell = [4.18, 4.72, 58.38, 89.44, 89.63, 75.85]

        if self.phil_params.calibrate_unit_cell.space_group is not None:
            self._spacegroup = self.phil_params.calibrate_unit_cell.space_group
        else:
            self._spacegroup = "P1"

        self._show_hkl = self.phil_params.calibrate_unit_cell.show_hkl

        self._cell_control_names = [
            "uc_a_ctrl",
            "uc_b_ctrl",
            "uc_c_ctrl",
            "uc_alpha_ctrl",
            "uc_beta_ctrl",
            "uc_gamma_ctrl",
        ]

        box = wx.BoxSizer(wx.HORIZONTAL)

        self.uc_a = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[0],
            value=self._cell[0],
        )
        box.Add(
            self.uc_a, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        box.Add(wx.StaticText(self, label="a"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_a)

        self.uc_alpha = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[3],
            value=self._cell[3],
        )
        box.Add(
            self.uc_alpha,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="alpha"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_alpha)

        sizer.Add(box)

        box = wx.BoxSizer(wx.HORIZONTAL)

        self.uc_b = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[1],
            value=self._cell[1],
        )
        box.Add(
            self.uc_b, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        box.Add(wx.StaticText(self, label="b"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_b)

        self.uc_beta = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[4],
            value=self._cell[4],
        )
        box.Add(
            self.uc_beta, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        box.Add(
            wx.StaticText(self, label="beta"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_beta)

        sizer.Add(box)

        box = wx.BoxSizer(wx.HORIZONTAL)

        self.uc_c = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[2],
            value=self._cell[2],
        )
        box.Add(
            self.uc_c, 0, wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL, 5
        )
        box.Add(wx.StaticText(self, label="c"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_c)

        self.uc_gamma = FloatSpin(
            self,
            digits=self.digits,
            name=self._cell_control_names[5],
            value=self._cell[5],
        )
        box.Add(
            self.uc_gamma,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="gamma"), 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCell, self.uc_gamma)

        sizer.Add(box)

        # Space group control
        box = wx.BoxSizer(wx.HORIZONTAL)

        self.space_group_ctrl = wx.TextCtrl(
            self, name="space group", value=self._spacegroup
        )
        box.Add(
            self.space_group_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Space group"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(wx.EVT_TEXT, self.OnSpaceGroup, self.space_group_ctrl)

        sizer.Add(box)

        # Distance control
        img = self.GetParent().GetParent()._img
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.distance_ctrl = FloatSpin(
            self,
            digits=self.digits,
            name="Detector Distance",
            value=img.get_detector_distance(),
        )
        self.distance_ctrl.SetIncrement(0.5)
        box.Add(
            self.distance_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )

        txtd = wx.StaticText(self, label="Detector Distance")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpin, self.distance_ctrl)
        sizer.Add(box)

        # Wavelength control
        img = self.GetParent().GetParent()._img
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.wavelength_ctrl = FloatSpin(
            self, digits=4, name="Wavelength", value=img.get_wavelength()
        )
        self.wavelength_ctrl.SetIncrement(0.05)
        box.Add(
            self.wavelength_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )

        txtw = wx.StaticText(self, label="Wavelength")
        box.Add(txtw, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpin, self.wavelength_ctrl)
        sizer.Add(box)

        # d_min control
        if self.phil_params.calibrate_unit_cell.d_min is not None:
            self.d_min = self.phil_params.calibrate_unit_cell.d_min
        else:
            self.d_min = 3.5
        box = wx.BoxSizer(wx.HORIZONTAL)
        self.d_min_ctrl = FloatSpin(
            self, digits=self.digits, name="d_min", value=self.d_min
        )
        box.Add(
            self.d_min_ctrl,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )

        txtd = wx.StaticText(self, label="Highest resolution for ring display")
        box.Add(txtd, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpin, self.d_min_ctrl)
        sizer.Add(box)

        # Centering controls.
        self._center = [0, 0]
        box = wx.BoxSizer(wx.HORIZONTAL)

        self.spinner_fast = FloatSpin(
            self, digits=self.digits, name="fast_ctrl", value=self._center[0]
        )
        box.Add(
            self.spinner_fast,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Center fast"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        self.Bind(EVT_FLOATSPIN, self.OnSpinCenter, self.spinner_fast)

        self.spinner_slow = FloatSpin(
            self, digits=self.digits, name="slow_ctrl", value=self._center[1]
        )
        box.Add(
            self.spinner_slow,
            0,
            wx.RIGHT | wx.TOP | wx.BOTTOM | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        box.Add(
            wx.StaticText(self, label="Center slow"),
            0,
            wx.ALL | wx.ALIGN_CENTER_VERTICAL,
            5,
        )
        sizer.Add(box)

        box = wx.BoxSizer(wx.HORIZONTAL)
        self.clear_button = wx.Button(self, -1, "Clear")
        box.Add(self.clear_button, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(wx.EVT_BUTTON, self.OnClear, self.clear_button)
        sizer.Add(box)

        origin_box = wx.BoxSizer(wx.HORIZONTAL)
        self.origin = wx.StaticText(self, label="")
        origin_box.Add(self.origin, 0, wx.ALL | wx.ALIGN_CENTER_VERTICAL, 5)
        self.Bind(EVT_FLOATSPIN, self.OnSpinCenter, self.spinner_slow)

        sizer.Add(origin_box)

        self.DrawRings()

    def __del__(self):
        if hasattr(self, "_ring_layer") and self._ring_layer is not None:
            self._pyslip.DeleteLayer(self._ring_layer)
            self._ring_layer = None

    def OnSpinCenter(self, event):
        obj = event.EventObject
        name = obj.GetName()

        if name == "fast_ctrl":
            self._center[0] = obj.GetValue()
        elif name == "slow_ctrl":
            self._center[1] = obj.GetValue()

        self.DrawRings()

    def OnSpinCell(self, event):
        obj = event.EventObject
        name = obj.GetName()

        self._cell[self._cell_control_names.index(name)] = obj.GetValue()

        self.DrawRings()

    def OnSpin(self, event):
        self.DrawRings()

    def OnSpaceGroup(self, event):
        obj = event.EventObject
        self._spacegroup = obj.GetValue()

        self.DrawRings()

    def OnClear(self, event):
        self.__del__()

    def _draw_rings_layer(self, dc, data, map_rel):
        """Draw a points layer.

        dc       the device context to draw on
        data     an iterable of point tuples:
                 (x, y, place, radius, colour, x_off, y_off, pdata)
        map_rel  points relative to map if True, MUST BE TRUE for lightweight
        Assumes all points are the same colour, saving 100's of ms.
        """

        assert map_rel is True
        if len(data) == 0:
            return
        (lon, lat, place, radius, colour, x_off, y_off, pdata) = data[0]

        scale = 2**self._pyslip.tiles.zoom_level

        # Draw points on map/view, using transparency if implemented.
        try:
            dc = wx.GCDC(dc)
        except NotImplementedError:
            pass
        dc.SetPen(wx.Pen(colour))
        dc.SetBrush(wx.Brush(colour, wx.TRANSPARENT))
        for (lon, lat, place, radius, colour, x_off, y_off, pdata) in data:
            (x, y) = self._pyslip.ConvertGeo2View((lon, lat))
            dc.DrawCircle(x, y, radius * scale)

    def DrawRings(self):
        frame = self.GetParent().GetParent()

        try:
            uc = symmetry(
                unit_cell=self._cell, space_group_symbol=str(self._spacegroup)
            )
            hkl_list = cctbx.miller.build_set(
                uc, False, d_min=self.d_min_ctrl.GetValue()
            )
        except Exception as e:
            frame.update_statusbar(str(e))
            return

        if self._show_hkl:
            hkl_list = hkl_list.common_set(
                cctbx.miller.set(
                    crystal_symmetry=uc, indices=self._show_hkl, anomalous_flag=False
                )
            )

        frame.update_statusbar(
            "%d %d %d %d %d %d, " % tuple(self._cell)
            + f"number of indices: {len(hkl_list.indices())}"
        )

        spacings = sorted(hkl_list.d_spacings(), key=lambda s: s[1], reverse=True)
        print(f"Printing spacings, len: {len(spacings)}")

        for d in spacings:
            print(d)

        detector = self._pyslip.tiles.raw_image.get_detector()

        wavelength = float(self.wavelength_ctrl.GetValue())
        distance = float(self.distance_ctrl.GetValue())
        pixel_size = detector[0].get_pixel_size()[
            0
        ]  # FIXME assumes square pixels, and that all panels use same pixel size

        twotheta = hkl_list.two_theta(wavelength=wavelength)
        L_mm = []
        L_pixels = []
        for tt in twotheta:
            L_mm.append(distance * math.tan(tt[1]))
        for lmm in L_mm:
            L_pixels.append(lmm / pixel_size)

        xrayframe = self.GetParent().GetParent()
        panel_id, beam_pixel_fast, beam_pixel_slow = xrayframe.get_beam_center_px()

        (
            beam_pixel_slow,
            beam_pixel_fast,
        ) = xrayframe.pyslip.tiles.flex_image.tile_readout_to_picture(
            panel_id, beam_pixel_slow - 0.5, beam_pixel_fast - 0.5
        )

        center = self._pyslip.tiles.picture_fast_slow_to_map_relative(
            beam_pixel_fast + self._center[0], beam_pixel_slow + self._center[1]
        )

        # XXX Transparency?
        ring_data = [
            (center[0], center[1], {"colour": "red", "radius": pxl}) for pxl in L_pixels
        ]

        # Remove the old ring layer, and draw a new one.
        if hasattr(self, "_ring_layer") and self._ring_layer is not None:
            self._pyslip.DeleteLayer(self._ring_layer)
            self._ring_layer = None
        self._ring_layer = self._pyslip.AddPointLayer(
            ring_data,
            map_rel=True,
            visible=True,
            show_levels=[-3, -2, -1, 0, 1, 2, 3, 4, 5],
            renderer=self._draw_rings_layer,
            name="<ring_layer>",
        )
        panel = detector[0]
        fast = col(panel.get_fast_axis())
        slow = col(panel.get_slow_axis())
        norm = col(panel.get_normal())
        x = -panel.pixel_to_millimeter(self._center)[0]
        y = -panel.pixel_to_millimeter(self._center)[1]
        z = -(panel.get_distance() - distance)
        origin = (fast * x + slow * y + norm * z) + col(panel.get_origin())
        self.origin.SetLabel("Panel 0 origin: %f, %f, %f" % origin.elems)



from __future__ import annotations

import copy
import os
import random

import pytest

from cctbx import crystal, sgtbx
from cctbx.sgtbx import bravais_types
from dxtbx.format import Format
from dxtbx.imageset import ImageSequence, ImageSetData
from dxtbx.model import (
    BeamFactory,
    Crystal,
    DetectorFactory,
    GoniometerFactory,
    ScanFactory,
)
from dxtbx.model.experiment_list import Experiment, ExperimentList
from dxtbx.serialize import load
from scitbx import matrix
from scitbx.math import euler_angles_as_matrix

from dials.algorithms.indexing.assign_indices import (
    AssignIndicesGlobal,
    AssignIndicesLocal,
)
from dials.array_family import flex


def random_rotation(angle_min=0, angle_max=360):
    angles = [random.uniform(angle_min, angle_max) for i in range(3)]
    print("Rotation: ", angles)
    return euler_angles_as_matrix(angles, deg=True)


@pytest.fixture
def crystal_factory():
    def _crystal_factory(space_group_symbol):
        # set random seeds so tests more reliable
        seed = 54321
        random.seed(seed)
        flex.set_random_seed(seed)

        space_group_info = sgtbx.space_group_info(symbol=space_group_symbol)
        space_group = space_group_info.group()
        unit_cell = space_group_info.any_compatible_unit_cell(
            volume=random.uniform(1e4, 1e6)
        )

        crystal_symmetry = crystal.symmetry(
            unit_cell=unit_cell, space_group=space_group
        )
        crystal_symmetry.show_summary()

        # the reciprocal matrix
        B = matrix.sqr(unit_cell.fractionalization_matrix()).transpose()
        U = random_rotation()
        A = U * B

        direct_matrix = A.inverse()
        return Crystal(
            direct_matrix[0:3],
            direct_matrix[3:6],
            direct_matrix[6:9],
            space_group=space_group,
        )

    return _crystal_factory


@pytest.fixture
def experiment():
    beam = BeamFactory.make_beam(wavelength=0.97625, sample_to_source=(0, 0, 1))

    detector = DetectorFactory.simple(
        sensor="PAD",
        distance=265.27,
        beam_centre=(210.7602, 205.27684),
        fast_direction="+x",
        slow_direction="-y",
        pixel_size=(0.172, 0.172),
        image_size=(2463, 2527),
        trusted_range=(0, 1e8),
    )

    goniometer = GoniometerFactory.single_axis()

    scan = ScanFactory.make_scan(
        image_range=(1, 20),
        exposure_times=0.067,
        oscillation=(82, 0.15),
        epochs=[0] * 20,
    )

    isetdata = ImageSetData(
        reader=Format.Reader(None, ["path"] * len(scan)), masker=None
    )
    iset = ImageSequence(
        isetdata, beam=beam, detector=detector, goniometer=goniometer, scan=scan
    )

    return Experiment(
        imageset=iset, beam=beam, detector=detector, goniometer=goniometer, scan=scan
    )


@pytest.mark.parametrize("space_group_symbol", bravais_types.acentric)
def test_assign_indices(space_group_symbol, experiment, crystal_factory):
    cryst_model = crystal_factory(space_group_symbol)
    experiment.crystal = cryst_model

    predicted_reflections = flex.reflection_table.from_predictions(experiment)
    use_fraction = 0.3
    use_sel = flex.random_selection(
        len(predicted_reflections), int(use_fraction * len(predicted_reflections))
    )
    predicted_reflections = predicted_reflections.select(use_sel)
    miller_indices = predicted_reflections["miller_index"]
    predicted_reflections["xyzobs.mm.value"] = predicted_reflections["xyzcal.mm"]
    predicted_reflections["id"] = flex.int(len(predicted_reflections), 0)
    predicted_reflections.map_centroids_to_reciprocal_space(
        ExperimentList([experiment])
    )

    # check that local and global indexing worked equally well in absence of errors
    result = CompareGlobalLocal(experiment, predicted_reflections, miller_indices)
    assert result.misindexed_local == 0
    assert result.misindexed_global == 0

    a, b, c = map(matrix.col, experiment.crystal.get_real_space_vectors())
    relative_error = 0.02
    a *= 1 + relative_error
    b *= 1 + relative_error
    c *= 1 + relative_error

    cryst_model2 = Crystal(a, b, c, space_group=cryst_model.get_space_group())
    experiment.crystal = cryst_model2

    result = CompareGlobalLocal(experiment, predicted_reflections, miller_indices)

    # check that the local indexing did a better job given the errors in the basis vectors
    # assert result.misindexed_local < result.misindexed_global
    assert result.misindexed_local == 0
    assert result.correct_local > result.correct_global
    # usually the number misindexed is much smaller than this
    assert result.misindexed_local < (0.001 * len(result.reflections_local))

    # the reciprocal matrix
    A = matrix.sqr(cryst_model.get_A())
    A = random_rotation(angle_max=0.5) * A

    direct_matrix = A.inverse()
    cryst_model2 = Crystal(
        direct_matrix[0:3],
        direct_matrix[3:6],
        direct_matrix[6:9],
        space_group=cryst_model.get_space_group(),
    )
    experiment.crystal = cryst_model2

    result = CompareGlobalLocal(experiment, predicted_reflections, miller_indices)

    # check that the local indexing did a better job given the errors in the basis vectors
    assert result.misindexed_local <= result.misindexed_global, (
        result.misindexed_local,
        result.misindexed_global,
    )
    assert result.misindexed_local < 0.01 * result.correct_local
    assert result.correct_local >= result.correct_global
    # usually the number misindexed is much smaller than this
    assert result.misindexed_local < (0.001 * len(result.reflections_local))


class CompareGlobalLocal:
    def __init__(self, experiment, reflections, expected_miller_indices):

        index_reflections_global = AssignIndicesGlobal()
        index_reflections_local = AssignIndicesLocal()

        # index reflections using simple "global" method
        self.reflections_global = copy.deepcopy(reflections)
        self.reflections_global["id"] = flex.int(len(self.reflections_global), -1)
        self.reflections_global["imageset_id"] = flex.int(
            len(self.reflections_global), 0
        )
        index_reflections_global(self.reflections_global, ExperimentList([experiment]))
        non_zero_sel = self.reflections_global["miller_index"] != (0, 0, 0)
        assert self.reflections_global["id"].select(~non_zero_sel).all_eq(-1)
        self.misindexed_global = (
            (expected_miller_indices == self.reflections_global["miller_index"])
            .select(non_zero_sel)
            .count(False)
        )
        self.correct_global = (
            expected_miller_indices == self.reflections_global["miller_index"]
        ).count(True)

        # index reflections using xds-style "local" method
        self.reflections_local = copy.deepcopy(reflections)
        self.reflections_local["id"] = flex.int(len(self.reflections_local), -1)
        index_reflections_local(self.reflections_local, ExperimentList([experiment]))
        non_zero_sel = self.reflections_local["miller_index"] != (0, 0, 0)
        assert self.reflections_local["id"].select(~non_zero_sel).all_eq(-1)
        self.misindexed_local = (
            (expected_miller_indices == self.reflections_local["miller_index"])
            .select(non_zero_sel)
            .count(False)
        )
        self.correct_local = (
            expected_miller_indices == self.reflections_local["miller_index"]
        ).count(True)

        print(
            "Global misindexed: %d, correct: %d, total: %d"
            % (
                self.misindexed_global,
                self.correct_global,
                len(self.reflections_global),
            )
        )
        print(
            " Local misindexed: %d, correct: %d, total: %d"
            % (self.misindexed_local, self.correct_local, len(self.reflections_local))
        )


def test_index_reflections(dials_regression):
    experiments_json = os.path.join(
        dials_regression, "indexing_test_data", "i04_weak_data", "experiments.json"
    )
    experiments = load.experiment_list(experiments_json, check_format=False)
    reflections = flex.reflection_table.from_file(
        os.path.join(
            dials_regression, "indexing_test_data", "i04_weak_data", "full.pickle"
        )
    )
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)
    reflections["imageset_id"] = flex.int(len(reflections), 0)
    reflections["id"] = flex.int(len(reflections), -1)
    AssignIndicesGlobal(tolerance=0.3)(reflections, experiments)
    assert "miller_index" in reflections
    counts = reflections["id"].counts()
    assert dict(counts) == {-1: 1390, 0: 114692}


def test_local_multiple_rotations(dials_data):
    """Test the fix for https://github.com/dials/dials/issues/1458"""

    experiments = load.experiment_list(
        dials_data("insulin_processed", pathlib=True) / "indexed.expt",
        check_format=False,
    )
    # Override the scan range to ensure we have 4 full rotations
    experiments[0].scan.set_image_range((1, 1440))

    # Generate some predicted reflections
    reflections = flex.reflection_table.from_predictions(experiments[0], dmin=4)
    reflections["imageset_id"] = flex.int(len(reflections), 0)
    reflections["id"] = flex.int(len(reflections), -1)
    reflections["xyzobs.px.value"] = reflections["xyzcal.px"]
    reflections["xyzobs.mm.value"] = reflections["xyzcal.mm"]
    predicted_miller_indices = reflections["miller_index"]

    # Prepare reflections for indexing
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)

    # Reset miller indices to ensure they are reindexed
    reflections["miller_index"] = flex.miller_index(len(reflections), (0, 0, 0))

    # Assign indices with the correct scan oscillation
    AssignIndicesLocal()(reflections, experiments)

    # Assert we have correctly indexed all reflections
    assert (reflections["miller_index"] == (0, 0, 0)).count(True) == 0
    assert (reflections["miller_index"] == predicted_miller_indices).count(False) == 0

    # Modify the scan oscillation such that we are out by 1 degree per rotation
    experiments[0].scan.set_oscillation((0, 1 - 1 / 360), deg=True)

    # Reset miller indices and re-map to reciprocal space
    reflections["miller_index"] = flex.miller_index(len(reflections), (0, 0, 0))
    reflections["id"] = flex.int(len(reflections), -1)
    reflections.centroid_px_to_mm(experiments)
    reflections.map_centroids_to_reciprocal_space(experiments)

    # Assign indices, this time with the incorrect scan oscillation
    AssignIndicesLocal()(reflections, experiments)

    # Assert that most reflections have been indexed
    indexed_sel = reflections["miller_index"] == (0, 0, 0)
    assert indexed_sel.count(True) < 10

    # Assert that all indexed miller indices are correct
    assert (
        (reflections["miller_index"] != predicted_miller_indices) & ~indexed_sel
    ).count(True) == 0



"""
Mask images to remove unwanted pixels.

This program is used to generate mask to specify which pixels should be
considered "invalid" during spot finding and integration. It provides a few
options to create simple masks using the detector trusted range, or from
simple shapes or by setting different resolution ranges.

Examples::

  dials.generate_mask models.expt border=5

  dials.generate_mask models.expt \\
    untrusted.rectangle=50,100,50,100 \\
    untrusted.circle=200,200,100

  dials.generate_mask models.expt d_max=2.00
"""


from __future__ import annotations

import logging
import os.path
import pickle
from typing import List, Optional, Tuple

import libtbx.phil as phil
from dxtbx.format.image import ImageBool
from dxtbx.model.experiment_list import ExperimentList
from scitbx.array_family import flex

import dials.util
import dials.util.log
import dials.util.masking
from dials.util.options import ArgumentParser, flatten_experiments

Masks = List[Tuple[flex.bool, ...]]

log = logging.getLogger("dials.generate_mask")

phil_scope = phil.parse(
    """
    output {
        mask = pixels.mask
            .type = path
            .help = "Name of output mask file."
        experiments = None
            .type = path
            .help = "Name of output experiment list file.  If this is set, a copy of "
                    "the experiments, modified with the generated pixel masks, "
                    "will be saved to this location."
        log = 'dials.generate_mask.log'
            .type = str
            .help = "The log filename."
    }

    include scope dials.util.masking.phil_scope
    """,
    process_includes=True,
)


def generate_mask(
    experiments: ExperimentList,
    params: phil.scope_extract,
) -> Tuple[Masks, Optional[ExperimentList]]:
    """
    Generate a pixel mask for each imageset in an experiment list.

    Use the masking parameters :param:`params` and the experiments in the experiment
    list :param:`experiments` to define a pixel mask for each of the associated
    imagesets.  The masks are generated using :mod:`dials.util.masking`.

    The masks will be saved to disk at the location specified by
    :attr:`params.output.mask`.  If the experiment list contains more than one
    imageset, multiple mask files will be produced, with filenames differentiated by
    an appended number.  Optionally, if a path :attr:`params.output.experiments` is
    set, a modified copy of :param:`experiments` with the masks applied will be saved
    to that location.

    Args:
        experiments: An experiment list containing only one imageset.
        params: Masking parameters, having the structure defined in
            :data:`phil_scope`.

    Returns:
        A list of masks, one for each imageset.

        A copy of :param:`experiments` with the masks applied (optional,
        only returned if :attr:`params.output.experiments` is set).
    """
    imagesets = experiments.imagesets()
    masks = []

    # Create output mask filenames
    num_imagesets = len(imagesets)
    if num_imagesets == 1:
        filenames = [params.output.mask]
    else:
        # If there is more than one imageset, append a number to each output filename
        name, ext = os.path.splitext(params.output.mask)
        pad = len(str(num_imagesets))
        filenames = [
            "{name}_{num:0{pad}}{ext}".format(name=name, num=i + 1, pad=pad, ext=ext)
            for i in range(num_imagesets)
        ]

    for imageset, filename in zip(imagesets, filenames):
        mask = dials.util.masking.generate_mask(imageset, params)
        masks.append(mask)

        # Save the mask to file
        log.info("Writing mask to %s", filename)
        with open(filename, "wb") as fh:
            pickle.dump(mask, fh)

        if params.output.experiments:
            # Apply the mask to the imageset
            imageset.external_lookup.mask.data = ImageBool(mask)
            imageset.external_lookup.mask.filename = filename

    if params.output.experiments:
        # Save the experiment list
        log.info("Saving experiments to %s", params.output.experiments)
        experiments.as_file(params.output.experiments)
    else:
        experiments = None

    return masks, experiments


@dials.util.show_mail_handle_errors()
def run(args: List[str] = None, phil: phil.scope = phil_scope) -> None:
    """
    Parse command-line arguments, run the script.

    Uses the DIALS option parser to extract an experiment list and
    parameters, then passes these to :func:`generate_mask`.

    Args:
        phil: PHIL scope for option parser.
        args: Arguments to parse. If None, :data:`sys.argv[1:]` will be used.
    """
    # Create the parser
    usage = "usage: dials.generate_mask [options] models.expt"
    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        epilog=__doc__,
        read_experiments=True,
        read_experiments_from_images=True,
    )

    # Parse the command line arguments
    params, options = parser.parse_args(args=args, show_diff_phil=True)
    experiments = flatten_experiments(params.input.experiments)

    # Configure logging
    dials.util.log.config(verbosity=options.verbose, logfile=params.output.log)

    # Check number of args
    if len(experiments) == 0:
        parser.print_help()
        return

    # Run the script
    generate_mask(experiments, params)


if __name__ == "__main__":
    run()



from __future__ import annotations

import copy
import os

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

from dials.algorithms.spot_prediction import ScanStaticReflectionPredictor
from dials.array_family import flex


class Data:
    def __init__(self, dials_data, dials_regression):
        self.experiments = ExperimentListFactory.from_json_file(
            dials_data("centroid_test_data", pathlib=True) / "experiments.json"
        )
        assert len(self.experiments) == 1
        self.experiments[0].imageset.set_beam(self.experiments[0].beam)
        self.experiments[0].imageset.set_detector(self.experiments[0].detector)
        self.experiments[0].imageset.set_goniometer(self.experiments[0].goniometer)
        self.experiments[0].imageset.set_scan(self.experiments[0].scan)

        reflection_filename = os.path.join(
            dials_regression, "prediction_test_data", "expected_reflections.pickle"
        )

        self.reflections = flex.reflection_table.from_file(reflection_filename)

    def predict_new(self, experiment=None, hkl=None, panel=None):
        if experiment is None:
            experiment = self.experiments[0]
        predict = ScanStaticReflectionPredictor(experiment)
        if hkl is None:
            return predict.for_ub(experiment.crystal.get_A())
        else:
            if panel is None:
                return predict(hkl)
            else:
                return predict(hkl, panel)


@pytest.fixture(scope="session")
def data(dials_data, dials_regression):
    return Data(dials_data, dials_regression)


def test_number_of_predictions(data):
    prediction_count = len(data.predict_new())
    assert prediction_count == 1996, "Predicted %d spots != 1996" % prediction_count

    modified_experiment = copy.deepcopy(data.experiments[0])
    modified_experiment.scan.set_oscillation((360, 0.2))
    prediction_count = len(data.predict_new(experiment=modified_experiment))
    assert prediction_count == 1996, (
        "Predicted %d spots != 1996 when rotated by 360 degrees" % prediction_count
    )


def test_vs_old(data):
    r_old = data.reflections
    r_new = data.predict_new()
    index1 = flex.size_t(
        sorted(range(len(r_old)), key=lambda x: r_old["miller_index"][x])
    )
    index2 = flex.size_t(
        sorted(range(len(r_new)), key=lambda x: r_new["miller_index"][x])
    )
    r_old = r_old.select(index1)
    r_new = r_new.select(index2)
    assert len(r_old) == len(r_new)
    for r1, r2 in zip(r_old.rows(), r_new.rows()):
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert r1["entering"] == r2["entering"]
        assert all(a == pytest.approx(b, abs=1e-7) for a, b in zip(r1["s1"], r2["s1"]))
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.px"], r2["xyzcal.px"])
        )
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.mm"], r2["xyzcal.mm"])
        )


def test_with_old_index_generator(data):
    predict = ScanStaticReflectionPredictor(data.experiments[0])
    r_old = predict.for_ub_old_index_generator(data.experiments[0].crystal.get_A())
    r_new = predict.for_ub(data.experiments[0].crystal.get_A())
    index1 = flex.size_t(
        sorted(range(len(r_old)), key=lambda x: r_old["miller_index"][x])
    )
    index2 = flex.size_t(
        sorted(range(len(r_new)), key=lambda x: r_new["miller_index"][x])
    )
    r_old = r_old.select(index1)
    r_new = r_new.select(index2)
    assert len(r_old) == len(r_new)
    for r1, r2 in zip(r_old.rows(), r_new.rows()):
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert r1["entering"] == r2["entering"]
        assert all(a == pytest.approx(b, abs=1e-7) for a, b in zip(r1["s1"], r2["s1"]))
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.px"], r2["xyzcal.px"])
        )
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.mm"], r2["xyzcal.mm"])
        )


def test_with_reflection_table(data):
    r_old = data.reflections
    r_new = flex.reflection_table()
    r_new["miller_index"] = r_old["miller_index"]
    r_new["panel"] = r_old["panel"]
    r_new["entering"] = r_old["entering"]
    predict = ScanStaticReflectionPredictor(data.experiments[0])
    predict.for_reflection_table(r_new, data.experiments[0].crystal.get_A())
    for r1, r2 in zip(r_old.rows(), r_new.rows()):
        assert r1["miller_index"] == r2["miller_index"]
        assert r1["panel"] == r2["panel"]
        assert r1["entering"] == r2["entering"]
        assert all(a == pytest.approx(b, abs=1e-7) for a, b in zip(r1["s1"], r2["s1"]))
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.px"], r2["xyzcal.px"])
        )
        assert all(
            a == pytest.approx(b, abs=1e-7)
            for a, b in zip(r1["xyzcal.mm"], r2["xyzcal.mm"])
        )

    # FIXME Fix the Spot prediction interface
    # def tst_with_hkl(self):
    # from dials.algorithms.spot_prediction import IndexGenerator

    # unit_cell = self.experiments[0].crystal.get_unit_cell()
    # space_group_type = self.experiments[0].crystal.get_space_group().type()
    # dmin = self.experiments[0].detector.get_max_resolution(
    # self.experiments[0].beam.get_s0())
    # indices = IndexGenerator(unit_cell, space_group_type, dmin)

    # r_old = self.predict_new()
    # r_new = self.predict_new(self.experiments[0].crystal.get_A(),indices.to_array())
    # assert(len(r_old) == len(r_new))
    # eps = 1e-7
    # for r1, r2 in zip(r_old.rows(), r_new.rows()):
    # assert(r1['miller_index'] == r2['miller_index'])
    # assert(r1['panel'] == r2['panel'])
    # assert(r1['entering'] == r2['entering'])
    # assert(all(abs(a-b) < eps for a, b in zip(r1['s1'], r2['s1'])))
    # assert(all(abs(a-b) < eps for a, b in zip(r1['xyzcal.px'], r2['xyzcal.px'])))
    # assert(all(abs(a-b) < eps for a, b in zip(r1['xyzcal.mm'], r2['xyzcal.mm'])))

    # def tst_with_hkl_and_panel(self):
    # from dials.algorithms.spot_prediction import IndexGenerator

    # unit_cell = self.experiments[0].crystal.get_unit_cell()
    # space_group_type = self.experiments[0].crystal.get_space_group().type()
    # dmin = self.experiments[0].detector.get_max_resolution(
    # self.experiments[0].beam.get_s0())
    # indices = IndexGenerator(unit_cell, space_group_type, dmin)
    # indices = indices.to_array()
    # try:
    # r_new = self.predict_new(indices, 1)
    # assert(False)
    # except Exception:
    # pass

    # r_old = self.predict_new()
    # r_new = self.predict_new(indices,0)
    # assert(len(r_old) < len(r_new))

    # def tst_with_hkl_and_panel_list(self):
    # from dials.algorithms.spot_prediction import IndexGenerator
    # from dials.array_family import flex
    # unit_cell = self.experiments[0].crystal.get_unit_cell()
    # space_group_type = self.experiments[0].crystal.get_space_group().type()
    # dmin = self.experiments[0].detector.get_max_resolution(
    # self.experiments[0].beam.get_s0())
    # indices = IndexGenerator(unit_cell, space_group_type, dmin)
    # indices = indices.to_array()

    # panels = flex.size_t(len(indices), 1)
    # try:
    # r_new = self.predict_new(indices, panels)
    # assert(False)
    # except Exception:
    # pass

    # panels = flex.size_t(len(indices), 0)
    # r_old = self.predict_new()
    # r_new = self.predict_new(indices, panels)
    # assert(len(r_old) < len(r_new))
